# SPDX-License-Identifier: Apache-2.0

from logging import getLogger
import numpy as np
from scipy.sparse import coo_matrix
from onnx import GraphProto
from ..proto import onnx_proto, TensorProto
from ..common.data_types import (
    guess_proto_type, _guess_numpy_type, _guess_type_proto_str,
    _guess_type_proto, FloatType, DoubleType, Int64Type, copy_type)
from ..common._topology import Variable
from ..common._registration import get_shape_calculator, get_converter


logger = getLogger('skl2onnx')


class GraphStateVar:
    pass


class GraphState:

    def __init__(self, inputs, output_names, operator_name, scope,
                 container, converter, onnx_prefix_name=None,
                 options=None, expected_inputs=None,
                 expected_outputs=None, input_range=None,
                 output_range=None, operator=None,
                 run_converters=False, input_types=None, **attrs):

        logger.debug(
            "[State] +%s n_inputs=%r n_outputs=%r",
            operator_name, -1 if inputs is None else len(inputs),
            -1 if output_names is None else len(output_names))
        self.inputs = inputs
        self._output_names = output_names
        self._input_range = input_range.copy() if input_range else [1, 1e9]
        self._output_range = output_range.copy() if output_range else [1, 1e9]
        self.scope = scope
        self.run_converters = run_converters
        self.operator = operator
        if hasattr(operator_name, 'fit'):
            from .. import get_model_alias
            self.operator_instance = operator_name
            self.is_model = True
            self.operator_name = get_model_alias(type(operator_name))
        elif operator_name.__class__.__name__ == "WrappedModelAlias":
            self.operator_instance = operator_name.model
            self.is_model = True
            self.operator_name = operator_name.alias
        else:
            self.operator_name = operator_name
            self.is_model = False
        self.container = container
        self.converter = converter
        self._expected_inputs = (
            None if expected_inputs is None else expected_inputs.copy())
        self._expected_outputs = (
            None if expected_outputs is None else expected_outputs.copy())
        self.computed_inputs_ = None
        self.computed_outputs_ = None
        self.sub_op_ = None
        self.onnx_prefix_name = onnx_prefix_name
        self.attrs = attrs
        self.options = options
        self.input_types = input_types

        for att in ['inputs', '_expected_inputs',
                    '_expected_outputs', 'computed_inputs_',
                    'computed_outputs_', '_outputs']:
            v = getattr(self, att, None)
            if v is None:
                continue
            if not isinstance(v, list):
                raise TypeError(
                    "Attribute %r must be a list not %r."
                    "" % (att, type(v)))
            for i, vi in enumerate(v):
                if hasattr(vi, 'state') or hasattr(vi, 'onx_op'):
                    continue
                if not isinstance(vi, (tuple, str, Variable, GraphStateVar)):
                    raise TypeError(
                        "Unexpected type %r for element %d of attribute %r "
                        "in %r." % (type(vi), i, att, v))
                if isinstance(vi, tuple) and len(vi) != 2:
                    raise ValueError(
                        "Unexpected value %r for element %d of attribute %r."
                        "" % (vi, i, att))
            change = []
            for vi in v:
                change.append((vi, None) if isinstance(vi, str) else vi)

        if self._output_names is not None:
            res = []
            if self._expected_outputs is not None:
                for i in range(0, len(self._expected_outputs)):
                    if i < len(self._output_names):
                        res.append(
                            (self._output_names[i],
                             self._expected_outputs[i][1]))
                    else:
                        res.append(self._expected_outputs[i])
            for i in range(len(res), len(self._output_names)):
                res.append((self._output_names[i], None))
            self._expected_outputs = res

        if self._expected_outputs is not None:
            res = []
            for p in self._expected_outputs:
                if isinstance(p[1], str) and p[1].startswith('tensor('):
                    res.append((p[0], _guess_type_proto_str(p[1], None)))
                else:
                    res.append(p)
            self._expected_outputs = res

        if self._expected_inputs is not None:
            res = []
            for p in self._expected_inputs:
                if isinstance(p[1], str) and p[1].startswith('tensor('):
                    res.append((p[0], _guess_type_proto_str(p[1], None)))
                else:
                    res.append(p)
            self._expected_inputs = res

    @property
    def onnx_prefix(self):
        if self.onnx_prefix_name is None:
            return self.operator_name
        return self.onnx_prefix_name + "_" + self.operator_name

    @property
    def outputs(self):
        self.run()
        return self.computed_outputs_

    def _get_var_name(self, var, in_out, operator=None, index=None):
        "input: True for output, False for input"
        if hasattr(var, 'add_to'):
            var.add_to(self.scope, self.container, operator=operator,
                       run_converters=self.run_converters)
            outputs = var.outputs
            if isinstance(outputs, list):
                vars = []
                for var in outputs:
                    if isinstance(var, (Variable, tuple)):
                        vars.append(var)
                    elif isinstance(var, str):
                        vars.append((var, None))
                if len(vars) == 0:
                    raise RuntimeError(
                        "Empty inputs outputs=%s var=%s in_out=%s "
                        "operator=%r." % (outputs, var, in_out, operator))
                return vars
            raise RuntimeError("Unexpected output type {}".format(outputs))

        def __fct__(var, operator):
            if isinstance(var, Variable):
                return [var]
            if isinstance(var, (np.ndarray, np.bool_, np.int64,
                                np.float32, np.float64,
                                np.int8, np.uint8)):
                return [self._add_constant(var)]
            if hasattr(var, 'ConstantValue'):
                return [
                    self._add_constant(var.ConstantValue, scope=self.scope)]
            if isinstance(var, str):
                return [(var, None)]
            if isinstance(var, tuple) and len(var) == 2:
                return [var]
            try:
                a, b = var
                return [(a, b)]
            except ValueError:
                pass
            raise RuntimeError("Unexpected type for parameter 'var': {0}."
                               "".format(type(var)))

        try:
            v = __fct__(var, operator)
        except TypeError as e:
            raise RuntimeError(
                "Unable to process one variable %s and operator=%s "
                "(name=%r)." % (var, operator, self.operator_name)) from e
        if v is None or not isinstance(v, list) or len(v) == 0:
            raise TypeError(
                "Unexpected type or empty value %r - %s." % (type(v), v))
        if in_out and self._output_names is not None and index is not None:
            if len(v) != 1:
                raise RuntimeError(
                    "Mismatch number of outputs between %s and %s." % (
                        v, self._output_names[index]))
            v2 = self.scope.get(var[0], None)
            if v2 is not None:
                v = [v2]
            try:
                vn = v[0][0]
            except IndexError as e:
                raise ValueError(
                    "Unexpected output %s in operator name %r."
                    "" % (vn, self.operator_name)) from e
            if (index >= len(self._output_names) and
                    index >= self._output_range[0]):
                return None
            try:
                vin = self._output_names[index]
            except IndexError as e:
                raise ValueError(
                    "Unexpected index %s in operator name %r with ."
                    "output names %s." % (
                        index, self.operator_name,
                        self._output_names)) from e
            if vn != vin:
                raise RuntimeError(
                    "Mismatched output name %r between %s and %s." % (
                        vn, v, vin))
        return v

    def _add_constant(self, cst, scope):

        def _ty_astype(cst):
            astype = cst.dtype
            try:
                ty = guess_proto_type(_guess_numpy_type(cst.dtype, cst.shape))
            except NotImplementedError as e:
                st = str(astype).lower()
                if st.startswith('u') or st.startswith("<u"):
                    ty = onnx_proto.TensorProto.STRING
                    astype = None
                    cst = np.array([s.encode('utf-8') for s in cst])
                else:
                    raise NotImplementedError(
                        "Unable to guess ONNX type from type {}. "
                        "You may raise an issue at https://github.com/onnx/"
                        "sklearn-onnx/issues.".format(
                            cst.dtype)) from e
            return cst, ty, astype

        if isinstance(cst, np.ndarray):
            shape = cst.shape
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            cst, ty, astype = _ty_astype(cst)
            if astype is not None:
                cst = cst.astype(astype)
            if ty == onnx_proto.TensorProto.STRING:
                value = [s.encode('utf-8') for s in cst.flatten()]
            else:
                value = cst.flatten()
            self.container.add_initializer(
                name, ty, shape, value)
            return (name, _guess_numpy_type(cst.dtype, cst.shape))

        if isinstance(cst, coo_matrix):
            shape = cst.shape
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            cst, ty, astype = _ty_astype(cst)
            self.container.add_initializer(
                name, ty, shape, cst.astype(astype))
            return (name, _guess_numpy_type(cst.dtype, cst.shape))

        if isinstance(cst, TensorProto):
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            self.container.add_initializer(name, None, None, cst)
            return (name, _guess_type_proto(cst, None))

        if isinstance(cst, np.int64):
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            ty = TensorProto.INT64
            self.container.add_initializer(name, ty, None, cst)
            return (name, Int64Type())

        if isinstance(cst, np.float32):
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            ty = TensorProto.FLOAT
            self.container.add_initializer(name, ty, None, float(cst))
            return (name, FloatType())

        if isinstance(cst, np.float64):
            name = self.scope.get_unique_variable_name(
                self.onnx_prefix + 'cst')
            ty = TensorProto.DOUBLE
            self.container.add_initializer(name, ty, None, float(cst))
            return (name, DoubleType())

        raise NotImplementedError(
            "Unable to add a constant of type {}. "
            "You may raise an issue at https://github.com/onnx/"
            "sklearn-onnx/issues.".format(type(cst)))

    @staticmethod
    def _get_output_name(output_names, output, scope):
        if isinstance(output, Variable):
            return output
        if isinstance(output, str):
            if output in output_names:
                return (output, None)
            return (scope.get_unique_variable_name(output), None)
        if isinstance(output, tuple):
            if output[0] in output_names:
                return output
            return (scope.get_unique_variable_name(output[0]), output[1])
        raise NotImplementedError(
            "Unexpected output type {} [{}]. "
            "You may raise an issue at https://github.com/onnx/"
            "sklearn-onnx/issues.".format(type(output), output))

    @staticmethod
    def _update_inputs(inputs, names, scope, expected_inputs,
                       input_range, input_types=None):
        new_inputs = []
        for inp in inputs:
            if isinstance(inp, (Variable, tuple, GraphStateVar)):
                new_inputs.append(inp)
                continue
            if hasattr(inp, 'get_output_type_inference'):
                etype = inp.get_output_type_inference(inputs)
                new_inputs.extend(etype)
                continue
            raise TypeError(
                "Unable to infer shape of inputs %r (type is %r)"
                "." % (inp, type(inp)))

        for i in range(0, len(new_inputs)):
            inp = new_inputs[i]
            if isinstance(inp, tuple) and len(inp) == 2:
                if input_types is not None and i < len(input_types):
                    stype = input_types[i]
                else:
                    stype = None if isinstance(inp[1], str) else inp[1]
                if scope is not None:
                    if inp[0] in scope.variables:
                        var = scope.variables[inp[0]]
                        if stype is not None:
                            var.check_compatible_type(stype)
                    else:
                        onnx_name = scope.get_unique_variable_name(inp[0])
                        var = Variable(
                            inp[0], onnx_name, type=stype, scope=scope)
                        scope.register_variable(var)
                else:
                    var = Variable(inp[0], inp[0], type=stype, scope=scope)
                new_inputs[i] = var
                inp = new_inputs[i]
            elif isinstance(inp, GraphStateVar):
                new_inputs[i] = inp.as_variable(scope)
                inp = new_inputs[i]
            elif not isinstance(inp, Variable):
                raise TypeError(
                    "Inputs %d - %r must be of type Variable." % (i, inp))

            if names is not None:
                try:
                    onnx_name = (
                        names[i] if isinstance(names[i], str)
                        else names[i][0])
                except IndexError as e:
                    raise IndexError(
                        "Wrong index %d, list=%s." % (i, names)) from e
                inp.set_onnx_name(onnx_name)

        # Second pass.
        if expected_inputs is not None:
            memo = {}
            for i, (name, ct) in enumerate(expected_inputs):
                if ct in memo:
                    memo[ct].append(i)
                else:
                    memo[ct] = [i]
            for i in range(0, len(new_inputs)):
                inp = new_inputs[i]
                if inp.type is None:
                    ct = expected_inputs[i][1]
                    if ct in memo:
                        for j in memo[ct]:
                            if (j >= len(new_inputs) and
                                    j >= input_range[0]):
                                continue
                            if new_inputs[j].type is not None:
                                new_inputs[i].set_type(
                                    new_inputs[j].type.__class__())
                                break

        # Overwrite types if input_types is specified.
        if input_types is not None:
            for i in range(len(new_inputs)):
                if i >= len(input_types):
                    raise RuntimeError(
                        "Mismatch between computed inputs[%d]=%r and "
                        "overwritten input_types[%d]=%r." % (
                            i, new_inputs, i, input_types))
                if input_types[i] is not None:
                    new_inputs[i].type = input_types[i]
        return new_inputs

    @staticmethod
    def _update_contraints(vars1, expected1, vars2, expected2, debug=None):
        memo = {}
        for va, ex in [(vars1, expected1), (vars2, expected2)]:
            if va is None or ex is None:
                continue
            for v, ct in zip(va, ex):
                if (isinstance(v, str) or (
                        hasattr(v, 'type') and v.type is None)):
                    continue
                vt = (copy_type(v.type)
                      if hasattr(v, 'type') else copy_type(v[1]))
                if isinstance(vt, str):
                    continue
                key = ct[1]
                if isinstance(key, str) and key[0] in ('T', 'I', 'V'):
                    if not isinstance(vt, str) and key not in memo:
                        memo[key] = []
                    memo[key].append(vt)

        for k, v in memo.items():
            if len(set(_.__class__ for _ in v)) != 1:
                raise RuntimeError(
                    "Conflicted constraint %r, got types %r operator=%s"
                    "." % (k, v, debug))
        for i in range(0, len(vars1)):
            inp = vars1[i]
            if isinstance(inp, str):
                continue
            if hasattr(inp, 'type') and inp.type is None:
                ct = expected1[i][1]
                if ct in memo:
                    vars1[i].set_type(copy_type(memo[ct][0]))
            elif isinstance(inp, tuple):
                ct = expected1[i][1]
                if ct in memo:
                    vars1[i] = (inp[0], copy_type(memo[ct][0]))

    def run(self):
        if self.computed_outputs_ is None:

            # We need to register all names in subgraphs and raise
            # an exception if the names are already taken.
            for k, v in self.attrs.items():
                if isinstance(v, GraphProto):
                    try:
                        self.scope.declare_existing_subgraph_name(v)
                    except NameError as e:
                        raise RuntimeError(
                            "A name exists both in the subgraph and "
                            "in the main graph. Use set_onnx_name_prefix to "
                            "to rename one of them, attribute=%r, "
                            "op_type=%r." % (
                                k, self.operator_name)) from e

            if self.operator is not None:
                expected_outputs = self.operator.outputs
            else:
                if self._expected_outputs is not None:
                    eoli = []
                    for i, o in enumerate(self._expected_outputs):
                        v = self._get_var_name(o, True, index=i)
                        if v is None:
                            continue
                        eoli.extend(v)
                    expected_outputs = eoli
                else:
                    expected_outputs = None

            logger.debug(
                "[State.run] id=%d op_name=%r is_model=%r "
                "expected_outputs=%r",
                id(self), self.operator_name, self.is_model, expected_outputs)

            inputs = []
            for i in self.inputs:
                v = self._get_var_name(i, False, index=None)
                inputs.extend(v)

            self.computed_inputs_ = GraphState._update_inputs(
                self.inputs, inputs, scope=self.scope,
                expected_inputs=self._expected_inputs,
                input_range=self._input_range,
                input_types=self.input_types)

            logger.debug(
                "[State.run] id=%d op_name=%r computed_inputs_=%r",
                id(self), self.operator_name, self.computed_inputs_)

            name = self.scope.get_unique_operator_name(self.onnx_prefix)
            if self.is_model:
                if self.sub_op_ is not None:
                    raise NotImplementedError(
                        "Attribute 'sub_op_' is not empty.")

                # a model is converted into a subgraph
                sub_op_inputs = self.computed_inputs_
                for v in sub_op_inputs:
                    if not isinstance(v, Variable):
                        raise TypeError(
                            "Every input variable must be a Variable not %r,"
                            " v=%r." % (type(v), v))
                    scope = v.scope
                    if hasattr(scope, 'variables'):
                        if v.onnx_name not in scope.variables:
                            raise RuntimeError(
                                "Variable %r missing from scope "
                                "(operator=%r, model=%r), list=%r." % (
                                    v, self.operator,
                                    type(self.operator_instance),
                                    list(sorted(self.scope.variables))))

                # output are not defined, we need to call a parser.
                from .._parse import _parse_sklearn
                self.scope.add_options(
                    id(self.operator_instance), self.options)
                try:
                    sub_outputs = _parse_sklearn(
                        self.scope, self.operator_instance, sub_op_inputs,
                        alias=self.operator_name)
                except RuntimeError as e:
                    raise RuntimeError(
                        "Unable to run parser for model type %r, inputs=%r "
                        "(input_types=%r)." % (
                            type(self.operator_instance), sub_op_inputs,
                            self.input_types)) from e
                set_input_names = set(v.onnx_name for v in sub_op_inputs)
                sub_op = None
                for op in self.scope.operators.values():
                    for inp in op.inputs:
                        if inp.onnx_name in set_input_names:
                            sub_op = op
                if (sub_outputs is None or
                        None in sub_outputs):
                    raise RuntimeError(
                        "Wrong result when parsing model {}.".format(
                            type(self.operator_instance)))

                # Checks operator outputs
                for out in sub_outputs:
                    if not isinstance(out, Variable):
                        raise TypeError(
                            "Output %s must be of type Variable." % out)
                self.sub_op_ = sub_op
                sub_op.outputs = sub_outputs

                shape_calc = get_shape_calculator(self.operator_name)
                logger.debug(
                    "[StateShape] call %r fed %r - %r", sub_op,
                    "".join(str(i.is_fed) for i in sub_op.inputs),
                    "".join(str(i.is_fed) for i in sub_op.outputs))
                shape_calc(sub_op)
                logger.debug("[StateShape] end - %r", sub_op)

                # Add Identity nodes to be consistent with `is_fed`
                # in Topology.
                if sub_op.outputs is not None and len(sub_op.outputs) > 0:
                    outputs = [
                        self.scope.declare_local_variable(
                            o.onnx_name, type=o.type)
                        for o in sub_op.outputs]
                elif (expected_outputs is not None and
                        len(expected_outputs) > 0):
                    outputs = [
                        self._get_output_name(
                            self._output_names, o, self.scope)
                        for o in expected_outputs]
                else:
                    raise RuntimeError(
                        "sub_op.outputs is None as well as expected_outputs "
                        "for operator %r." % sub_op)

                if len(outputs) != len(sub_op.outputs):
                    raise RuntimeError(
                        "Mismatched number of outputs %s and %s." % (
                            outputs, sub_op.outputs))

                for i, out in enumerate(sub_op.outputs):
                    var = outputs[i]
                    self.container.add_node(
                        'Identity', [out.onnx_name], [var[0]],
                        name=self.scope.get_unique_operator_name("SubOpId"))
                self.computed_outputs_ = outputs
                self.computed_inputs2_ = sub_op.inputs
                self.computed_outputs2_ = [
                    (v[0], v[1]) for v in self.computed_outputs_]

                if self.run_converters:
                    # The parser was run on sub-operators but not the
                    # converter.
                    conv = get_converter(self.operator_name)
                    logger.debug(
                        "[StateConv] %r fed %r - %r", sub_op,
                        "".join(str(i.is_fed) for i in sub_op.inputs),
                        "".join(str(i.is_fed) for i in sub_op.outputs))
                    conv(self.scope, sub_op, self.container)
                    logger.debug("[StateConv] %r - end.", sub_op)
                else:
                    if (expected_outputs is not None and
                            len(sub_op.outputs) == len(expected_outputs)):
                        for v1, v2 in zip(sub_op.outputs, expected_outputs):
                            if isinstance(v2, tuple):
                                v2 = v2[0]
                            if (hasattr(v1, 'onnx_name') and
                                    hasattr(v2, 'onnx_name')):
                                if v1.onnx_name != v2.onnx_name:
                                    # One identity is missing
                                    n = self.scope.get_unique_operator_name(
                                        'idgstate')
                                    self.container.add_node(
                                        'Identity', [v1.onnx_name],
                                        [v2.onnx_name], name=n)
            else:

                def _name_(obj):
                    if isinstance(obj, tuple) and len(obj) == 2:
                        return obj[0]
                    if hasattr(obj, 'onnx_name'):
                        return obj.onnx_name
                    raise TypeError(
                        "Unable to extract variable name from %r." % obj)

                # only one node is added
                if self.options is not None:
                    raise RuntimeError(
                        "Options must be empty for node %r but is it %r." % (
                            self.operator_name, self.options))
                outputs = [
                    self._get_output_name(self._output_names, o, self.scope)
                    for o in expected_outputs]
                input_names = [_name_(i) for i in inputs]
                output_names = [_name_(i) for i in outputs]
                self.container.add_node(
                    self.operator_name, input_names, output_names,
                    name=name, **self.attrs)

                computed_outputs = [
                    (name, ct[1]) for name, ct in zip(
                        output_names, self._expected_outputs)]
                self._update_contraints(
                    computed_outputs, self._expected_outputs,
                    self.computed_inputs_, self._expected_inputs,
                    debug=self.operator_name)

                # Registers the variables into scope.
                self.computed_outputs_ = []
                for name, kind in computed_outputs:
                    if isinstance(kind, str):
                        self.computed_outputs_.append((name, kind))
                    else:
                        var = self.scope.declare_local_variable(
                            name, kind, missing_type=True)
                        # name already comes from
                        # scope.get_unique_variable_name
                        var.set_onnx_name(name)
                        var.init_status(is_fed=True)
                        self.computed_outputs_.append(var)

            logger.debug('[State.run] end id=%d', id(self))



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import (
    FloatTensorType, Int64TensorType, DoubleTensorType)
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_truncated_svd_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, K]

    Transform feature dimension from C to K
    """
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[
            FloatTensorType, Int64TensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType])

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Only 2-D tensor(s) can be input(s).')

    cls_type = operator.inputs[0].type.__class__
    if cls_type != DoubleTensorType:
        cls_type = FloatTensorType
    N = operator.inputs[0].get_first_dimension()
    K = (operator.raw_operator.n_components
         if operator.type == 'SklearnTruncatedSVD'
         else operator.raw_operator.n_components_)

    operator.outputs[0].type = cls_type([N, K])


register_shape_calculator('SklearnIncrementalPCA',
                          calculate_sklearn_truncated_svd_output_shapes)
register_shape_calculator('SklearnPCA',
                          calculate_sklearn_truncated_svd_output_shapes)
register_shape_calculator('SklearnTruncatedSVD',
                          calculate_sklearn_truncated_svd_output_shapes)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from scipy.special import digamma
from sklearn.mixture import BayesianGaussianMixture, GaussianMixture
try:
    from sklearn.mixture._gaussian_mixture import _compute_log_det_cholesky
except ImportError:
    # scikit-learn < 0.22
    from sklearn.mixture.gaussian_mixture import _compute_log_det_cholesky
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type
from ..algebra.onnx_ops import (
    OnnxAdd, OnnxSub, OnnxMul, OnnxGemm, OnnxReduceSumSquareApi18,
    OnnxReduceLogSumExpApi18, OnnxExp, OnnxArgMax, OnnxConcat,
    OnnxReduceSumApi11, OnnxLog, OnnxReduceMaxApi18, OnnxEqual, OnnxCast
)
from ..proto import onnx_proto


def _estimate_log_gaussian_prob(X, means, precisions_chol,
                                covariance_type, dtype, op_version,
                                combined_reducesum):
    """
    Converts the same function into ONNX.
    Returns log probabilities.
    """
    n_components = means.shape[0]
    n_features = means.shape[1]
    opv = op_version

    # self._estimate_log_prob(X)
    log_det = _compute_log_det_cholesky(
        precisions_chol, covariance_type, n_features).astype(
            dtype)

    if covariance_type == 'full':
        # shape(op.means_) = (n_components, n_features)
        # shape(op.precisions_cholesky_) =
        #   (n_components, n_features, n_features)

        # log_prob = np.empty((n_samples, n_components))
        # for k, (mu, prec_chol) in enumerate(zip(means, precisions_chol)):
        #     y = np.dot(X, prec_chol) - np.dot(mu, prec_chol)
        #     log_prob[:, k] = np.sum(np.square(y), axis=1)

        ys = []
        for c in range(n_components):
            prec_chol = precisions_chol[c, :, :]
            cst = - np.dot(means[c, :], prec_chol)
            y = OnnxGemm(X, prec_chol.astype(dtype),
                         cst.astype(dtype), alpha=1.,
                         beta=1., op_version=opv)
            if combined_reducesum:
                y2s = OnnxReduceSumApi11(OnnxMul(y, y, op_version=opv),
                                         axes=[1], op_version=opv)
            else:
                y2s = OnnxReduceSumSquareApi18(y, axes=[1], op_version=opv)
            ys.append(y2s)
        log_prob = OnnxConcat(*ys, axis=1, op_version=opv)

    elif covariance_type == 'tied':
        # shape(op.means_) = (n_components, n_features)
        # shape(op.precisions_cholesky_) =
        #   (n_features, n_features)

        # log_prob = np.empty((n_samples, n_components))
        # for k, mu in enumerate(means):
        #     y = np.dot(X, precisions_chol) - np.dot(mu, precisions_chol)
        #     log_prob[:, k] = np.sum(np.square(y), axis=1)

        ys = []
        for f in range(n_components):
            cst = - np.dot(means[f, :], precisions_chol)
            y = OnnxGemm(X, precisions_chol.astype(dtype),
                         cst.astype(dtype),
                         alpha=1., beta=1., op_version=opv)
            if combined_reducesum:
                y2s = OnnxReduceSumApi11(OnnxMul(y, y, op_version=opv),
                                         axes=[1], op_version=opv)
            else:
                y2s = OnnxReduceSumSquareApi18(y, axes=[1], op_version=opv)
            ys.append(y2s)
        log_prob = OnnxConcat(*ys, axis=1, op_version=opv)

    elif covariance_type == 'diag':
        # shape(op.means_) = (n_components, n_features)
        # shape(op.precisions_cholesky_) =
        #   (n_components, n_features)

        # precisions = precisions_chol ** 2
        # log_prob = (np.sum((means ** 2 * precisions), 1) -
        #             2. * np.dot(X, (means * precisions).T) +
        #             np.dot(X ** 2, precisions.T))

        precisions = (precisions_chol ** 2).astype(dtype)
        mp = np.sum((means ** 2 * precisions), 1).astype(dtype)
        zeros = np.zeros((n_components, ), dtype=dtype)
        xmp = OnnxGemm(
            X, (means * precisions).T.astype(dtype),
            zeros, alpha=-2., beta=0., op_version=opv)
        term = OnnxGemm(OnnxMul(X, X, op_version=opv),
                        precisions.T.astype(dtype),
                        zeros, alpha=1., beta=0., op_version=opv)
        log_prob = OnnxAdd(
            OnnxAdd(mp.astype(dtype), xmp, op_version=opv),
            term, op_version=opv)

    elif covariance_type == 'spherical':
        # shape(op.means_) = (n_components, n_features)
        # shape(op.precisions_cholesky_) = (n_components, )

        # precisions = precisions_chol ** 2
        # log_prob = (np.sum(means ** 2, 1) * precisions -
        #             2 * np.dot(X, means.T * precisions) +
        #             np.outer(row_norms(X, squared=True), precisions))

        zeros = np.zeros((n_components, ), dtype=dtype)
        precisions = (precisions_chol ** 2).astype(dtype)
        if combined_reducesum:
            normX = OnnxReduceSumApi11(OnnxMul(X, X, op_version=opv),
                                       axes=[1], op_version=opv)
        else:
            normX = OnnxReduceSumSquareApi18(X, axes=[1], op_version=opv)
        outer = OnnxGemm(
            normX, precisions[np.newaxis, :].astype(dtype),
            zeros.astype(dtype), alpha=1., beta=1., op_version=opv)
        xmp = OnnxGemm(
            X, (means.T * precisions).astype(dtype),
            zeros, alpha=-2., beta=0., op_version=opv)
        mp = (np.sum(means ** 2, 1) * precisions).astype(dtype)
        log_prob = OnnxAdd(mp, OnnxAdd(xmp, outer, op_version=opv),
                           op_version=opv)
    else:
        raise RuntimeError("Unknown op.covariance_type='{}'. Upgrade "
                           "to a more recent version of skearn-onnx "
                           "or raise an issue.".format(covariance_type))
    # -.5 * (cst + log_prob) + log_det
    cst = np.array([n_features * np.log(2 * np.pi)]).astype(dtype)
    add = OnnxAdd(cst, log_prob, op_version=opv)
    mul = OnnxMul(add, np.array([-0.5], dtype=dtype),
                  op_version=opv)
    if isinstance(log_det, (np.float32, np.float64, float)):
        log_det = np.array([log_det], dtype=dtype)

    return OnnxAdd(mul, log_det.astype(dtype), op_version=opv)


def convert_sklearn_gaussian_mixture(scope: Scope, operator: Operator,
                                     container: ModelComponentContainer):
    """
    Converter for *GaussianMixture*,
    *BayesianGaussianMixture*.
    """
    X = operator.inputs[0]
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    elif operator.target_opset < 11:
        raise RuntimeError(
            "Some needed operators are not available below opset 11"
            " to convert model %r" % type(operator.raw_operator))
    out = operator.outputs
    op = operator.raw_operator
    n_components = op.means_.shape[0]
    opv = container.target_opset
    options = container.get_options(op, dict(score_samples=None))
    add_score = options.get('score_samples', False)
    combined_reducesum = not container.is_allowed(
        {'ReduceLogSumExp', 'ReduceSumSquare'})
    if add_score and len(out) != 3:
        raise RuntimeError("3 outputs are expected.")

    if X.type is not None:
        if X.type.shape[1] != op.means_.shape[1]:
            raise RuntimeError(
                "Dimension mismath between expected number of features {} "
                "and ONNX graphs expectations {}.".format(
                    op.means_.shape[1], X.type.shape[1]))
    n_features = op.means_.shape[1]

    # All comments come from scikit-learn code and tells
    # which functions is being onnxified.
    # def _estimate_weighted_log_prob(self, X):
    log_weights = op._estimate_log_weights().astype(dtype)

    log_gauss = _estimate_log_gaussian_prob(
        X, op.means_, op.precisions_cholesky_, op.covariance_type,
        dtype, opv, combined_reducesum)

    if isinstance(op, BayesianGaussianMixture):
        # log_gauss = (_estimate_log_gaussian_prob(
        #   X, self.means_, self.precisions_cholesky_, self.covariance_type) -
        #   .5 * n_features * np.log(self.degrees_of_freedom_))

        log_lambda = n_features * np.log(2.) + np.sum(digamma(
            .5 * (op.degrees_of_freedom_ -
                  np.arange(0, n_features)[:, np.newaxis])), 0)
        cst_log_lambda = .5 * (log_lambda - n_features / op.mean_precision_)
        cst = cst_log_lambda - .5 * n_features * np.log(op.degrees_of_freedom_)
        if isinstance(cst, np.ndarray):
            cst_array = cst.astype(dtype)
        else:
            cst_array = np.array([cst], dtype=dtype)
        log_gauss = OnnxAdd(log_gauss, cst_array, op_version=opv)
    elif not isinstance(op, GaussianMixture):
        raise RuntimeError(
            "The converter does not support type {}.".format(
                type(op)))

    # self._estimate_log_prob(X) + self._estimate_log_weights()
    weighted_log_prob = OnnxAdd(log_gauss, log_weights, op_version=opv)

    # labels
    if container.is_allowed('ArgMax'):
        labels = OnnxArgMax(weighted_log_prob, axis=1,
                            output_names=out[:1], op_version=opv)
    else:
        mxlabels = OnnxReduceMaxApi18(
            weighted_log_prob, axes=[1], op_version=opv)
        zeros = OnnxEqual(
            OnnxSub(weighted_log_prob, mxlabels, op_version=opv),
            np.array([0], dtype=dtype),
            op_version=opv)
        toint = OnnxCast(zeros, to=onnx_proto.TensorProto.INT64,
                         op_version=opv)
        mulind = OnnxMul(toint,
                         np.arange(n_components).astype(np.int64),
                         op_version=opv)
        labels = OnnxReduceMaxApi18(
            mulind, axes=[1], output_names=out[:1], op_version=opv)

    # def _estimate_log_prob_resp():
    # np.exp(log_resp)
    # weighted_log_prob = self._estimate_weighted_log_prob(X)
    # log_prob_norm = logsumexp(weighted_log_prob, axis=1)
    # with np.errstate(under='ignore'):
    #    log_resp = weighted_log_prob - log_prob_norm[:, np.newaxis]
    if add_score:
        outnames = out[2:3]
    else:
        outnames = None

    if combined_reducesum:
        max_weight = OnnxReduceMaxApi18(
            weighted_log_prob, axes=[1], op_version=opv)
        log_prob_norm_demax = OnnxLog(
            OnnxReduceSumApi11(
                OnnxExp(
                    OnnxSub(weighted_log_prob, max_weight, op_version=opv),
                    op_version=opv),
                axes=[1], op_version=opv),
            op_version=opv)
        log_prob_norm = OnnxAdd(log_prob_norm_demax, max_weight,
                                op_version=opv, output_names=outnames)
    else:
        log_prob_norm = OnnxReduceLogSumExpApi18(
            weighted_log_prob, axes=[1], op_version=opv,
            output_names=outnames)
    log_resp = OnnxSub(weighted_log_prob, log_prob_norm, op_version=opv)

    # probabilities
    probs = OnnxExp(log_resp, output_names=out[1:2], op_version=opv)

    # final
    labels.add_to(scope, container)
    probs.add_to(scope, container)
    if add_score:
        log_prob_norm.add_to(scope, container)


register_converter('SklearnGaussianMixture', convert_sklearn_gaussian_mixture,
                   options={'score_samples': [True, False]})
register_converter('SklearnBayesianGaussianMixture',
                   convert_sklearn_gaussian_mixture,
                   options={'score_samples': [True, False]})



# SPDX-License-Identifier: Apache-2.0

import textwrap
import warnings
from types import MethodType
import numpy
from numpy.testing import assert_almost_equal
try:
    from scipy.sparse import csr_matrix
except ImportError:
    from scipy.sparse.csr import csr_matrix
from sklearn.base import TransformerMixin, ClassifierMixin
from sklearn.base import RegressorMixin, BaseEstimator
from sklearn.pipeline import Pipeline, FeatureUnion
try:
    from sklearn.compose import ColumnTransformer, TransformedTargetRegressor
except ImportError:
    # not avaiable in 0.19
    ColumnTransformer = None
    TransformedTargetRegressor = None


def enumerate_pipeline_models(pipe, coor=None, vs=None):
    """
    Enumerates all the models within a pipeline.
    """
    if coor is None:
        coor = (0,)
    yield coor, pipe, vs
    if hasattr(pipe, 'transformer_and_mapper_list') and len(
            pipe.transformer_and_mapper_list):
        # azureml DataTransformer
        raise NotImplementedError("Unable to handle this specific case.")
    elif hasattr(pipe, 'mapper') and pipe.mapper:
        # azureml DataTransformer
        for couple in enumerate_pipeline_models(pipe.mapper, coor + (0,)):
            yield couple
    elif hasattr(pipe, 'built_features'):
        # sklearn_pandas.dataframe_mapper.DataFrameMapper
        for i, (columns, transformers, _) in enumerate(pipe.built_features):
            if isinstance(columns, str):
                columns = (columns,)
            if transformers is None:
                yield (coor + (i,)), None, columns
            else:
                for couple in enumerate_pipeline_models(transformers,
                                                        coor + (i,),
                                                        columns):
                    yield couple
    elif isinstance(pipe, Pipeline):
        for i, (_, model) in enumerate(pipe.steps):
            for couple in enumerate_pipeline_models(model, coor + (i,)):
                yield couple
    elif ColumnTransformer is not None and isinstance(pipe, ColumnTransformer):
        for i, (_, fitted_transformer, column) in enumerate(pipe.transformers):
            for couple in enumerate_pipeline_models(
                    fitted_transformer, coor + (i,), column):
                yield couple
    elif isinstance(pipe, FeatureUnion):
        for i, (_, model) in enumerate(pipe.transformer_list):
            for couple in enumerate_pipeline_models(model, coor + (i,)):
                yield couple
    elif TransformedTargetRegressor is not None and isinstance(
            pipe, TransformedTargetRegressor):
        raise NotImplementedError(
            "Not yet implemented for TransformedTargetRegressor.")
    elif isinstance(pipe, (TransformerMixin, ClassifierMixin, RegressorMixin)):
        pass
    elif isinstance(pipe, BaseEstimator):
        pass
    else:
        raise TypeError(
            "Parameter pipe is not a scikit-learn object: {}\n{}".format(
                type(pipe), pipe))


class BaseEstimatorDebugInformation:
    """
    Stores information when the outputs of a pipeline
    is computed. It as added by function
    :func:`_alter_model_for_debugging`.
    """

    def __init__(self, model):
        self.model = model
        self.inputs = {}
        self.outputs = {}
        self.methods = {}
        if hasattr(model, "transform") and callable(model.transform):
            model._debug_transform = model.transform
            self.methods["transform"] = \
                lambda model, X: model._debug_transform(X)
        if hasattr(model, "predict") and callable(model.predict):
            model._debug_predict = model.predict
            self.methods["predict"] = lambda model, X: model._debug_predict(X)
        if hasattr(model, "predict_proba") and callable(model.predict_proba):
            model._debug_predict_proba = model.predict_proba
            self.methods["predict_proba"] = \
                lambda model, X: model._debug_predict_proba(X)
        if hasattr(model, "decision_function") and callable(
            model.decision_function):  # noqa
            model._debug_decision_function = model.decision_function  # noqa
            self.methods["decision_function"] = \
                lambda model, X: model._debug_decision_function(X)

    def __repr__(self):
        """
        usual
        """
        return self.to_str()

    def to_str(self, nrows=5):
        """
        Tries to produce a readable message.
        """
        rows = ['BaseEstimatorDebugInformation({})'.format(
            self.model.__class__.__name__)]
        for k in sorted(self.inputs):
            if k in self.outputs:
                rows.append('  ' + k + '(')
                self.display(self.inputs[k], nrows)
                rows.append(textwrap.indent(
                    self.display(self.inputs[k], nrows), '   '))
                rows.append('  ) -> (')
                rows.append(textwrap.indent(
                    self.display(self.outputs[k], nrows), '   '))
                rows.append('  )')
            else:
                raise KeyError(
                    "Unable to find output for method '{}'.".format(k))
        return "\n".join(rows)

    def display(self, data, nrows):
        """
        Displays the first
        """
        text = str(data)
        rows = text.split('\n')
        if len(rows) > nrows:
            rows = rows[:nrows]
            rows.append('...')
        if hasattr(data, 'shape'):
            rows.insert(0, "shape={}".format(data.shape))
        return "\n".join(rows)


def _alter_model_for_debugging(skl_model, recursive=False):
    """
    Overwrite methods transform, predict or predict_proba
    to collect the last inputs and outputs
    seen in these methods.

    :param skl_model: *scikit-learn* pipeline or model
    :param recursive: alter the current model (False) or git into
        contained models
    """

    def transform(self, X, *args, **kwargs):
        self._debug.inputs['transform'] = X
        y = self._debug.methods['transform'](self, X, *args, **kwargs)
        self._debug.outputs['transform'] = y
        return y

    def predict(self, X, *args, **kwargs):
        self._debug.inputs['predict'] = X
        y = self._debug.methods['predict'](self, X, *args, **kwargs)
        self._debug.outputs['predict'] = y
        return y

    def predict_proba(self, X, *args, **kwargs):
        self._debug.inputs['predict_proba'] = X
        y = self._debug.methods['predict_proba'](self, X, *args, **kwargs)
        self._debug.outputs['predict_proba'] = y
        return y

    def decision_function(self, X, *args, **kwargs):
        self._debug.inputs['decision_function'] = X
        y = self._debug.methods['decision_function'](self, X, *args, **kwargs)
        self._debug.outputs['decision_function'] = y
        return y

    new_methods = {
        'decision_function': decision_function,
        'transform': transform,
        'predict': predict,
        'predict_proba': predict_proba,
    }

    if hasattr(skl_model, '_debug'):
        raise RuntimeError("The same operator cannot be used twice in "
                           "the same pipeline or this method was called "
                           "a second time.")

    if recursive:
        for model_ in enumerate_pipeline_models(skl_model):
            model = model_[1]
            model._debug = BaseEstimatorDebugInformation(model)
            for k in model._debug.methods:
                try:
                    setattr(model, k, MethodType(new_methods[k], model))
                except AttributeError:
                    warnings.warn("Unable to overwrite method '{}' for class "
                                  "{}.".format(k, type(model)))
    else:
        skl_model._debug = BaseEstimatorDebugInformation(skl_model)
        for k in skl_model._debug.methods:
            try:
                setattr(skl_model, k, MethodType(new_methods[k], skl_model))
            except AttributeError:
                warnings.warn("Unable to overwrite method '{}' for class "
                              "{}.".format(k, type(skl_model)))


def collect_intermediate_steps(model, *args, **kwargs):
    """
    Converts a scikit-learn model into ONNX with :func:`convert_sklearn`
    and returns intermediate results for each included operator.

    :param model: model or pipeline to convert
    :param args: arguments for :func:`convert_sklearn`
    :param kwargs: optional arguments for :func:`convert_sklearn`

    The model *model* is modified by the function,
    it should be pickled first to be retrieved unaltered.
    This function is used to check every intermediate model in
    a pipeline.
    """
    if 'intermediate' in kwargs:
        if not kwargs['intermediate']:
            raise ValueError("Parameter intermediate must be true.")
        del kwargs['intermediate']

    from .. import convert_sklearn
    from ..helpers.onnx_helper import select_model_inputs_outputs
    from ..common import MissingShapeCalculator, MissingConverter
    try:
        model_onnx, topology = convert_sklearn(
            model, *args, intermediate=True, **kwargs)
    except (MissingShapeCalculator, MissingConverter):
        # The model cannot be converted.
        raise

    steps = []
    for operator in topology.unordered_operator_iterator():
        if operator.raw_operator is None:
            continue
        _alter_model_for_debugging(operator.raw_operator)
        inputs = [i.full_name for i in operator.inputs]
        outputs = [o.full_name for o in operator.outputs]
        steps.append({
            'model': operator.raw_operator,
            'model_onnx': model_onnx,
            'inputs': inputs,
            'outputs': outputs,
            'onnx_step': select_model_inputs_outputs(
                model_onnx, outputs=outputs)
        })
    return steps


def compare_objects(o1, o2, decimal=4):
    """
    Compares two objects assuming they are vectors or matrices.
    *o1* and *o2* can be a numpy array, a sparse matrix,
    a dataframe. The function raises an exception if it cannot
    convert both object into the same type or the comparison
    fails.

    :param o1: a dataframe, a series, an array a sparse matrix
    :param o2: a dataframe, a series, an array a sparse matrix
    :param decimal: parameter decimal for assert_almost_equal
    """

    def convert(o):
        if isinstance(o, list) and len(o) == 1:
            if isinstance(o[0], numpy.ndarray):
                if o[0].dtype in (numpy.str_, object, str):
                    o = list(o[0])
                else:
                    o = o[0]
        # Following line avoid importing pandas and taking
        # dependency on pandas.
        if o.__class__.__name__ == "Series":
            c = list(o)
        elif isinstance(o, numpy.ndarray):
            c = o
        elif isinstance(o, csr_matrix):
            c = o.todense()
        elif isinstance(o, list):
            c = o.copy()
        elif isinstance(o, tuple):
            c = list(o)
        else:
            raise TypeError("Unexpected type {}.".format(type(o)))
        return c

    def to_string(c):
        s = str(c)
        if len(s) > 200:
            s = s[:200] + "..."
        return s

    c1 = convert(o1)
    c2 = convert(o2)
    reason = None
    if isinstance(c2, list) and isinstance(c2[0], dict):
        res = numpy.zeros((len(c2), max(len(c) for c in c2)))
        for i, row in enumerate(c2):
            for k, v in row.items():
                res[i, k] = v
        c2 = res
    if isinstance(c1, numpy.ndarray) and isinstance(c2, list):
        c1 = list(c1.ravel())
    if isinstance(c1, list) and isinstance(c2, list):
        try:
            res = c1 == c2
            reason = 'list-equal'
        except ValueError:  # lgtm [py/unreachable-statement]
            res = False
            reason = 'list'
    elif isinstance(c1, numpy.ndarray) and isinstance(c2, numpy.ndarray):
        try:
            assert_almost_equal(c1, c2, decimal=decimal)
            res = True
        except (AssertionError, TypeError):
            reason = 'array'
            cc1 = c1.ravel()
            cc2 = c2.ravel()
            try:
                assert_almost_equal(cc1, cc2, decimal=decimal)
                res = True
            except (AssertionError, TypeError) as e:
                res = False
                reason = 'array-ravel' + str(e)
    else:
        raise TypeError("Types {} and {}".format(type(c1), type(c2)))
    if not res:
        msg = "o1 and o2 are different ({})\n---o1---\n{}\n---o2---\n{}"
        raise ValueError(msg.format(reason, to_string(c1), to_string(c2)))



# SPDX-License-Identifier: Apache-2.0

from sklearn.base import is_classifier
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .._parse import _parse_sklearn


def convert_pipeline(scope: Scope, operator: Operator,
                     container: ModelComponentContainer):
    model = operator.raw_operator
    inputs = operator.inputs
    for step in model.steps:
        step_model = step[1]
        if is_classifier(step_model):
            scope.add_options(id(step_model), options={'zipmap': False})
            container.add_options(id(step_model), options={'zipmap': False})
        outputs = _parse_sklearn(scope, step_model, inputs,
                                 custom_parsers=None)
        inputs = outputs
    if len(outputs) != len(operator.outputs):
        raise RuntimeError(
            "Mismatch between pipeline output %d and "
            "last step outputs %d." % (
                len(outputs), len(operator.outputs)))
    for fr, to in zip(outputs, operator.outputs):
        container.add_node(
            'Identity', fr.full_name, to.full_name,
            name=scope.get_unique_operator_name("Id" + operator.onnx_name))


def convert_feature_union(scope: Scope, operator: Operator,
                          container: ModelComponentContainer):
    raise NotImplementedError(
        "This converter not needed so far. It is usually handled "
        "during parsing.")


def convert_column_transformer(scope: Scope, operator: Operator,
                               container: ModelComponentContainer):
    raise NotImplementedError(
        "This converter not needed so far. It is usually handled "
        "during parsing.")


register_converter('SklearnPipeline', convert_pipeline)
register_converter('SklearnFeatureUnion', convert_feature_union)
register_converter('SklearnColumnTransformer', convert_column_transformer)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.shape_calculator import (
    _calculate_linear_classifier_output_shapes)


def voting_classifier_shape_calculator(operator):
    return _calculate_linear_classifier_output_shapes(
        operator, enable_type_checking=False)


register_shape_calculator(
    'SklearnVotingClassifier', voting_classifier_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..common.data_types import guess_proto_type
from ..common._apply_operation import (
    apply_add, apply_cast, apply_concat, apply_identity,
    apply_reshape, apply_sub)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def _forward_pass(scope, container, model, activations, proto_dtype):
    """
    Perform a forward pass on the network by computing the values of
    the neurons in the hidden layers and the output layer.
    """
    activations_map = {
        'identity': 'Identity', 'tanh': 'Tanh', 'logistic': 'Sigmoid',
        'relu': 'Relu', 'softmax': 'Softmax'
    }

    out_activation_result_name = scope.get_unique_variable_name(
        'out_activations_result')

    # Iterate over the hidden layers
    for i in range(model.n_layers_ - 1):
        coefficient_name = scope.get_unique_variable_name('coefficient')
        intercepts_name = scope.get_unique_variable_name('intercepts')
        mul_result_name = scope.get_unique_variable_name('mul_result')
        add_result_name = scope.get_unique_variable_name('add_result')

        container.add_initializer(
            coefficient_name, proto_dtype,
            model.coefs_[i].shape, model.coefs_[i].ravel())
        container.add_initializer(
            intercepts_name, proto_dtype,
            [1, len(model.intercepts_[i])], model.intercepts_[i])

        container.add_node(
            'MatMul', [activations[i], coefficient_name],
            mul_result_name, name=scope.get_unique_operator_name('MatMul'))
        apply_add(scope, [mul_result_name, intercepts_name],
                  add_result_name, container, broadcast=1)

        # For the hidden layers
        if (i + 1) != (model.n_layers_ - 1):
            activations_result_name = scope.get_unique_variable_name(
                'next_activations')

            container.add_node(
                activations_map[model.activation], add_result_name,
                activations_result_name,
                name=scope.get_unique_operator_name(
                    activations_map[model.activation]))
            activations.append(activations_result_name)

    # For the last layer
    container.add_node(
        activations_map[model.out_activation_], add_result_name,
        out_activation_result_name,
        name=scope.get_unique_operator_name(activations_map[model.activation]))
    activations.append(out_activation_result_name)

    return activations


def _predict(scope, input_name, container, model, proto_dtype):
    """
    This function initialises the input layer, calls _forward_pass()
    and returns the final layer.
    """
    cast_input_name = scope.get_unique_variable_name('cast_input')

    apply_cast(scope, input_name, cast_input_name,
               container, to=proto_dtype)

    # forward propagate
    activations = _forward_pass(scope, container, model, [cast_input_name],
                                proto_dtype)
    return activations[-1]


def convert_sklearn_mlp_classifier(scope: Scope, operator: Operator,
                                   container: ModelComponentContainer):
    """
    Converter for MLPClassifier.
    This function calls _predict() which returns the probability scores
    of the positive class in case of binary labels and class
    probabilities in case of multi-class. It then calculates probability
    scores for the negative class in case of binary labels. It
    calculates the class labels and sets the output.
    """
    mlp_op = operator.raw_operator
    classes = mlp_op.classes_
    class_type = onnx_proto.TensorProto.STRING

    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    y_pred = _predict(scope, operator.inputs[0].full_name, container, mlp_op,
                      proto_dtype)

    if (np.issubdtype(mlp_op.classes_.dtype, np.floating) or
            mlp_op.classes_.dtype == np.bool_):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(mlp_op.classes_.dtype, np.integer):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    if len(classes) == 2:
        unity_name = scope.get_unique_variable_name('unity')
        negative_class_proba_name = scope.get_unique_variable_name(
            'negative_class_proba')
        container.add_initializer(unity_name, proto_dtype, [], [1])
        apply_sub(scope, [unity_name, y_pred],
                  negative_class_proba_name, container, broadcast=1)
        apply_concat(scope, [negative_class_proba_name, y_pred],
                     operator.outputs[1].full_name, container, axis=1)
    else:
        apply_identity(scope, y_pred,
                       operator.outputs[1].full_name, container)

    if mlp_op._label_binarizer.y_type_ == 'multilabel-indicator':
        binariser_output_name = scope.get_unique_variable_name(
            'binariser_output')

        container.add_node('Binarizer', y_pred, binariser_output_name,
                           threshold=0.5, op_domain='ai.onnx.ml')
        apply_cast(
            scope, binariser_output_name, operator.outputs[0].full_name,
            container, to=onnx_proto.TensorProto.INT64)
    else:
        classes_name = scope.get_unique_variable_name('classes')
        container.add_initializer(classes_name, class_type,
                                  classes.shape, classes)

        container.add_node('ArgMax', operator.outputs[1].full_name,
                           argmax_output_name, axis=1,
                           name=scope.get_unique_operator_name('ArgMax'))
        container.add_node(
            'ArrayFeatureExtractor', [classes_name, argmax_output_name],
            array_feature_extractor_result_name, op_domain='ai.onnx.ml',
            name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

        if class_type == onnx_proto.TensorProto.INT32:
            reshaped_result_name = scope.get_unique_variable_name(
                'reshaped_result')

            apply_reshape(scope, array_feature_extractor_result_name,
                          reshaped_result_name, container,
                          desired_shape=(-1,))
            apply_cast(
                scope, reshaped_result_name, operator.outputs[0].full_name,
                container, to=onnx_proto.TensorProto.INT64)
        else:
            apply_reshape(scope, array_feature_extractor_result_name,
                          operator.outputs[0].full_name, container,
                          desired_shape=(-1,))


def convert_sklearn_mlp_regressor(scope: Scope, operator: Operator,
                                  container: ModelComponentContainer):
    """
    Converter for MLPRegressor.
    This function calls _predict() which returns the scores.
    """
    mlp_op = operator.raw_operator

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    y_pred = _predict(scope, operator.inputs[0].full_name, container, mlp_op,
                      proto_dtype=proto_dtype)
    apply_reshape(scope, y_pred, operator.output_full_names,
                  container, desired_shape=(-1, 1))


register_converter('SklearnMLPClassifier',
                   convert_sklearn_mlp_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnMLPRegressor',
                   convert_sklearn_mlp_regressor)



# SPDX-License-Identifier: Apache-2.0

import packaging.version as pv
import numpy as np
from onnx.helper import make_tensor
from sklearn import __version__
from ..proto import onnx_proto
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import (
    apply_add, apply_cast, apply_clip, apply_concat, apply_div, apply_exp,
    apply_mul, apply_reshape, apply_sub, apply_topk, apply_transpose
)
from ..common.data_types import (
    FloatTensorType, DoubleTensorType, guess_proto_type, guess_numpy_type,
    Int64TensorType)
from ..common._registration import register_converter
from .._supported_operators import sklearn_operator_name_map


def _scikit_learn_before_022():
    if '.dev' in __version__:
        return pv.Version(
            __version__.split(".dev")[0]) < pv.Version("0.22")
    if '.post' in __version__:
        return pv.Version(
            __version__.split(".post")[0]) < pv.Version("0.22")
    return pv.Version(__version__) < pv.Version("0.22")


def _samme_proba(scope, container, proba_name, weight,
                 zero_name, classes_ind_name, one_name):
    weight_name = scope.get_unique_variable_name('weight')
    container.add_initializer(
        weight_name, onnx_proto.TensorProto.FLOAT, [], [weight])

    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    container.add_node('ArgMax', proba_name,
                       argmax_output_name,
                       name=scope.get_unique_operator_name('ArgMax'),
                       axis=1)
    equal_name = scope.get_unique_variable_name('equal')
    container.add_node('Equal', [argmax_output_name, classes_ind_name],
                       equal_name,
                       name=scope.get_unique_operator_name('Equal'))

    max_proba_name = scope.get_unique_variable_name('probsmax')
    container.add_node('Where', [equal_name, one_name, zero_name],
                       max_proba_name,
                       name=scope.get_unique_operator_name('Where'))

    samme_proba_name = scope.get_unique_variable_name('samme_proba')
    apply_mul(scope, [max_proba_name, weight_name],
              samme_proba_name, container, broadcast=1)
    return samme_proba_name


def _samme_r_proba(scope, container, proba_name, n_classes, dtype, pdtype):
    clipped_proba_name = scope.get_unique_variable_name('clipped_proba')
    log_proba_name = scope.get_unique_variable_name('log_proba')
    reduced_proba_name = scope.get_unique_variable_name('reduced_proba')
    reshaped_result_name = scope.get_unique_variable_name('reshaped_result')
    inverted_n_classes_name = scope.get_unique_variable_name(
        'inverted_n_classes')
    n_classes_minus_one_name = scope.get_unique_variable_name(
        'n_classes_minus_one')
    prod_result_name = scope.get_unique_variable_name('prod_result')
    sub_result_name = scope.get_unique_variable_name('sub_result')
    samme_proba_name = scope.get_unique_variable_name('samme_proba')

    container.add_initializer(
        inverted_n_classes_name, pdtype, [], [1. / n_classes])
    container.add_initializer(
        n_classes_minus_one_name, pdtype, [], [n_classes - 1])

    try:
        cst_min = np.finfo(np.float64).eps.astype(dtype)
    except TypeError:
        raise TypeError("Unable to convert {} (type {}) into {}.".format(
            np.finfo(float).eps, type(np.finfo(float).eps), dtype))

    apply_clip(
        scope, proba_name, clipped_proba_name, container,
        operator_name=scope.get_unique_operator_name('ClipAda'),
        min=dtype(cst_min))
    container.add_node(
        'Log', clipped_proba_name, log_proba_name,
        name=scope.get_unique_operator_name('Log'))
    if container.target_opset < 13:
        container.add_node(
            'ReduceSum', log_proba_name, reduced_proba_name, axes=[1],
            name=scope.get_unique_operator_name('ReduceSum'))
    else:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            'ReduceSum', [log_proba_name, axis_name], reduced_proba_name,
            name=scope.get_unique_operator_name('ReduceSum'))
    apply_reshape(scope, reduced_proba_name,
                  reshaped_result_name, container,
                  desired_shape=(-1, 1))
    apply_mul(scope, [reshaped_result_name, inverted_n_classes_name],
              prod_result_name, container, broadcast=1)
    apply_sub(scope, [log_proba_name, prod_result_name],
              sub_result_name, container, broadcast=1)
    apply_mul(scope, [sub_result_name, n_classes_minus_one_name],
              samme_proba_name, container, broadcast=1)
    return samme_proba_name


def _normalise_probability(scope, container, operator, proba_names_list,
                           model):
    est_weights_sum_name = scope.get_unique_variable_name('est_weights_sum')
    summation_prob_name = scope.get_unique_variable_name('summation_prob')
    div_result_name = scope.get_unique_variable_name('div_result')
    exp_operand_name = scope.get_unique_variable_name('exp_operand')
    exp_result_name = scope.get_unique_variable_name('exp_result')
    reduced_exp_result_name = scope.get_unique_variable_name(
        'reduced_exp_result')
    normaliser_name = scope.get_unique_variable_name('normaliser')
    zero_scalar_name = scope.get_unique_variable_name('zero_scalar')
    comparison_result_name = scope.get_unique_variable_name(
        'comparison_result')
    cast_output_name = scope.get_unique_variable_name('cast_output')
    zero_filtered_normaliser_name = scope.get_unique_variable_name(
        'zero_filtered_normaliser')
    mul_operand_name = scope.get_unique_variable_name('mul_operand')
    cast_normaliser_name = scope.get_unique_variable_name('cast_normaliser')

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    container.add_initializer(
        est_weights_sum_name, proto_dtype,
        [], [model.estimator_weights_.sum()])
    container.add_initializer(
        mul_operand_name, proto_dtype,
        [], [1. / (model.n_classes_ - 1)])
    container.add_initializer(zero_scalar_name,
                              onnx_proto.TensorProto.INT32, [], [0])

    container.add_node('Sum', proba_names_list,
                       summation_prob_name,
                       name=scope.get_unique_operator_name('Sum'))
    apply_div(scope, [summation_prob_name, est_weights_sum_name],
              div_result_name, container, broadcast=1)
    apply_mul(scope, [div_result_name, mul_operand_name],
              exp_operand_name, container, broadcast=1)
    apply_exp(scope, exp_operand_name, exp_result_name, container)
    if container.target_opset < 13:
        container.add_node(
            'ReduceSum', exp_result_name, reduced_exp_result_name, axes=[1],
            name=scope.get_unique_operator_name('ReduceSum'))
    else:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            'ReduceSum', [exp_result_name, axis_name], reduced_exp_result_name,
            name=scope.get_unique_operator_name('ReduceSum'))
    apply_reshape(scope, reduced_exp_result_name,
                  normaliser_name, container,
                  desired_shape=(-1, 1))
    apply_cast(scope, normaliser_name, cast_normaliser_name,
               container, to=onnx_proto.TensorProto.INT32)
    container.add_node('Equal', [cast_normaliser_name, zero_scalar_name],
                       comparison_result_name,
                       name=scope.get_unique_operator_name('Equal'))
    apply_cast(scope, comparison_result_name, cast_output_name,
               container, to=proto_dtype)
    apply_add(scope, [normaliser_name, cast_output_name],
              zero_filtered_normaliser_name,
              container, broadcast=0)
    apply_div(scope, [exp_result_name, zero_filtered_normaliser_name],
              operator.outputs[1].full_name, container, broadcast=1)
    return operator.outputs[1].full_name


def _generate_raw_scores(scope, container, operator, proba_names_list, model):
    summation_prob_name = scope.get_unique_variable_name('summation_proba')
    est_weights_sum_name = scope.get_unique_variable_name('est_weights')

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    container.add_initializer(
        est_weights_sum_name, proto_dtype,
        [], [model.estimator_weights_.sum()])

    container.add_node(
        'Sum', proba_names_list, summation_prob_name,
        name=scope.get_unique_operator_name('Sum'))

    if len(model.classes_) == 2:
        div_res_name = scope.get_unique_variable_name('div_res')
        operand_name = scope.get_unique_variable_name('operand')
        neg_name = scope.get_unique_variable_name('neg')
        mul_res_name = scope.get_unique_variable_name('mul_res')
        pos_class_scores_name = scope.get_unique_variable_name(
            'pos_class_scores')
        neg_class_scores_name = scope.get_unique_variable_name(
            'neg_class_scores')
        container.add_initializer(
            operand_name, proto_dtype,
            [2], [-1, 1])
        container.add_initializer(
            neg_name, proto_dtype,
            [], [-1])

        apply_div(scope, [summation_prob_name, est_weights_sum_name],
                  div_res_name, container, broadcast=1)
        apply_mul(scope, [div_res_name, operand_name],
                  mul_res_name, container, broadcast=1)
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', mul_res_name, pos_class_scores_name, axes=[1],
                name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [mul_res_name, axis_name], pos_class_scores_name,
                name=scope.get_unique_operator_name('ReduceSum'))
        apply_mul(scope, [pos_class_scores_name, neg_name],
                  neg_class_scores_name, container, broadcast=1)
        apply_concat(
            scope, [neg_class_scores_name, pos_class_scores_name],
            operator.outputs[1].full_name, container, axis=1)
    else:
        apply_div(scope, [summation_prob_name, est_weights_sum_name],
                  operator.outputs[1].full_name, container, broadcast=1)
    return operator.outputs[1].full_name


def convert_sklearn_ada_boost_classifier(scope: Scope, operator: Operator,
                                         container: ModelComponentContainer):
    """
    Converter for AdaBoost classifier.
    This function goes through the list of estimators and uses
    TreeEnsembleClassifer op to calculate class probabilities
    for each estimator. Then it calculates the weighted sum
    across all the estimators depending on the algorithm
    picked during trainging (SAMME.R or SAMME) and normalises
    the probability score for the final result. Label is
    calculated by simply doing an argmax of the probability scores.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))['nocl']:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__))
    op = operator.raw_operator
    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']
    classes = op.classes_
    class_type = onnx_proto.TensorProto.STRING
    if np.issubdtype(classes.dtype, np.floating):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype('int')
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode('utf-8') for s in classes])

    argmax_output_name = scope.get_unique_variable_name('argmax_output')
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        'array_feature_extractor_result')

    classes_name = scope.get_unique_variable_name('classes')
    container.add_initializer(classes_name, class_type, classes.shape, classes)

    proba_names_list = []
    classes_ind_name = None
    zero_name = None
    one_name = None
    classes_ind_name = None

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    proba_type = operator.inputs[0].type.__class__
    add_cast = proto_dtype == onnx_proto.TensorProto.INT64
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    for i_est, estimator in enumerate(op.estimators_):
        label_name = scope.declare_local_variable(
            'elab_name_%d' % i_est, Int64TensorType())
        proba_name = scope.declare_local_variable(
            'eprob_name_%d' % i_est, proba_type())

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = operator.inputs
        this_operator.outputs.extend([label_name, proba_name])

        if add_cast:
            this_operator = scope.declare_local_operator('SklearnCast')
            this_operator.inputs.append(proba_name)
            var_name = scope.declare_local_variable('cast', FloatTensorType())
            this_operator.outputs.append(var_name)
            proba_name = var_name

        if op.algorithm == 'SAMME.R':
            cur_proba_name = _samme_r_proba(
                scope, container, proba_name.onnx_name, len(classes),
                dtype, proto_dtype)
        else:
            # SAMME
            if _scikit_learn_before_022() and not use_raw_scores:
                weight_name = scope.get_unique_variable_name('weight')
                samme_proba_name = scope.get_unique_variable_name(
                    'samme_proba')
                container.add_initializer(
                    weight_name, onnx_proto.TensorProto.FLOAT,
                    [], [op.estimator_weights_[i_est]])
                apply_mul(scope, [proba_name.onnx_name, weight_name],
                          samme_proba_name, container, broadcast=1)
                cur_proba_name = samme_proba_name
            else:
                if classes_ind_name is None:
                    classes_ind_name = scope.get_unique_variable_name(
                        'classes_ind3')
                    container.add_initializer(
                        classes_ind_name, onnx_proto.TensorProto.INT64,
                        (1, len(classes)), list(range(len(classes))))

                if zero_name is None:
                    shape_name = scope.get_unique_variable_name('shape')
                    container.add_node(
                        'Shape', proba_name.onnx_name, shape_name,
                        name=scope.get_unique_operator_name('Shape'))

                    zero_name = scope.get_unique_variable_name('zero')
                    container.add_node(
                        'ConstantOfShape', shape_name, zero_name,
                        name=scope.get_unique_operator_name('CoSA'),
                        value=make_tensor(
                            "value", onnx_proto.TensorProto.FLOAT,
                            (1, ), [0]))

                    one_name = scope.get_unique_variable_name('one')
                    container.add_node(
                        'ConstantOfShape', shape_name, one_name,
                        name=scope.get_unique_operator_name('CoSB'),
                        value=make_tensor(
                            "value", onnx_proto.TensorProto.FLOAT,
                            (1, ), [1.]))

                cur_proba_name = _samme_proba(
                    scope, container, proba_name.onnx_name,
                    op.estimator_weights_[i_est], zero_name,
                    classes_ind_name, one_name)

        proba_names_list.append(cur_proba_name)

    function = (_generate_raw_scores if use_raw_scores
                else _normalise_probability)
    class_prob_name = function(scope, container, operator,
                               proba_names_list, op)
    container.add_node('ArgMax', class_prob_name,
                       argmax_output_name,
                       name=scope.get_unique_operator_name('ArgMax'), axis=1)
    container.add_node(
        'ArrayFeatureExtractor', [classes_name, argmax_output_name],
        array_feature_extractor_result_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))

    if class_type == onnx_proto.TensorProto.INT32:
        reshaped_result_name = scope.get_unique_variable_name(
            'reshaped_result')

        apply_reshape(scope, array_feature_extractor_result_name,
                      reshaped_result_name, container,
                      desired_shape=(-1,))
        apply_cast(scope, reshaped_result_name, operator.outputs[0].full_name,
                   container, to=onnx_proto.TensorProto.INT64)
    else:
        apply_reshape(scope, array_feature_extractor_result_name,
                      operator.outputs[0].full_name, container,
                      desired_shape=(-1,))


def _get_estimators_label(scope, operator, container, model):
    """
    This function computes labels for each estimator and returns
    a tensor produced by concatenating the labels.
    """
    if isinstance(operator.inputs[0].type, DoubleTensorType):
        var_type = DoubleTensorType
    else:
        var_type = FloatTensorType
    concatenated_labels_name = scope.get_unique_variable_name(
        'concatenated_labels')

    input_name = operator.inputs
    estimators_results_list = []
    for i, estimator in enumerate(model.estimators_):
        estimator_label_name = scope.declare_local_variable(
            'est_label_%d' % i, var_type([None, 1]))

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = input_name
        this_operator.outputs.append(estimator_label_name)

        estimators_results_list.append(estimator_label_name.onnx_name)

    apply_concat(scope, estimators_results_list, concatenated_labels_name,
                 container, axis=1)
    return concatenated_labels_name


def cum_sum(scope, container, rnn_input_name, sequence_length, proto_dtype):
    opv = container.target_opset
    weights_cdf_name = scope.get_unique_variable_name('weights_cdf')
    if opv < 11:
        transposed_input_name = scope.get_unique_variable_name(
            'transposed_input')
        reshaped_result_name = scope.get_unique_variable_name(
            'reshaped_result')
        weights_name = scope.get_unique_variable_name('weights')
        rec_weights_name = scope.get_unique_variable_name('rec_weights')
        rnn_output_name = scope.get_unique_variable_name('rnn_output')
        permuted_rnn_y_name = scope.get_unique_variable_name('permuted_rnn_y')

        container.add_initializer(weights_name,
                                  proto_dtype, [1, 1, 1], [1])
        container.add_initializer(rec_weights_name,
                                  proto_dtype, [1, 1, 1], [1])

        apply_transpose(scope, rnn_input_name, transposed_input_name,
                        container, perm=(1, 0))
        apply_reshape(scope, transposed_input_name, reshaped_result_name,
                      container, desired_shape=(sequence_length, -1, 1))
        container.add_node(
            'RNN', inputs=[reshaped_result_name,
                           weights_name, rec_weights_name],
            outputs=[rnn_output_name], activations=['Affine'],
            name=scope.get_unique_operator_name('RNN'),
            activation_alpha=[1.0], activation_beta=[0.0], hidden_size=1)
        apply_transpose(scope, rnn_output_name, permuted_rnn_y_name, container,
                        perm=(2, 0, 1, 3))
        apply_reshape(
            scope, permuted_rnn_y_name, weights_cdf_name, container,
            desired_shape=(-1, sequence_length))
    else:
        axis_name = scope.get_unique_variable_name('axis_name')
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT32,
                                  [], [1])
        container.add_node(
            'CumSum', [rnn_input_name, axis_name], [weights_cdf_name],
            name=scope.get_unique_operator_name('CumSum'),
            op_version=11)
    return weights_cdf_name


def _apply_gather_elements(scope, container, inputs, output, axis,
                           dim, zero_type, suffix):
    if container.target_opset >= 11:
        container.add_node(
            'GatherElements', inputs, output, op_version=11, axis=axis,
            name=scope.get_unique_operator_name('GatEls' + suffix))
    else:
        classes_ind_name = scope.get_unique_variable_name('classes_ind2')
        container.add_initializer(
            classes_ind_name, onnx_proto.TensorProto.INT64,
            (1, dim), list(range(dim)))

        shape_name = scope.get_unique_variable_name('shape')
        container.add_node(
            'Shape', inputs[0], shape_name,
            name=scope.get_unique_operator_name('Shape'))
        zero_name = scope.get_unique_variable_name('zero')
        zero_val = (0 if zero_type == onnx_proto.TensorProto.INT64
                    else 0.)
        container.add_node(
            'ConstantOfShape', shape_name, zero_name,
            name=scope.get_unique_operator_name('CoSA'),
            value=make_tensor("value", zero_type,
                              (1, ), [zero_val]), op_version=9)

        equal_name = scope.get_unique_variable_name('equal')
        container.add_node('Equal', [inputs[1], classes_ind_name],
                           equal_name,
                           name=scope.get_unique_operator_name('Equal'))

        selected = scope.get_unique_variable_name('selected')
        container.add_node('Where', [equal_name, inputs[0], zero_name],
                           selected,
                           name=scope.get_unique_operator_name('Where'))
        if container.target_opset < 13:
            container.add_node(
                'ReduceSum', selected, output, axes=[1],
                name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                'ReduceSum', [selected, axis_name], output,
                name=scope.get_unique_operator_name('ReduceSum'))


def convert_sklearn_ada_boost_regressor(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    """
    Converter for AdaBoost regressor.
    This function first calls _get_estimators_label() which returns a
    tensor of concatenated labels predicted by each estimator. Then,
    median is calculated and returned as the final output.

    Note: This function creates an ONNX model which can predict on only
    one instance at a time because ArrayFeatureExtractor can only
    extract based on the last axis, so we can't fetch different columns
    for different rows.
    """
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    op = operator.raw_operator

    negate_name = scope.get_unique_variable_name('negate')
    estimators_weights_name = scope.get_unique_variable_name(
        'estimators_weights')
    half_scalar_name = scope.get_unique_variable_name('half_scalar')
    last_index_name = scope.get_unique_variable_name('last_index')
    negated_labels_name = scope.get_unique_variable_name('negated_labels')
    sorted_values_name = scope.get_unique_variable_name('sorted_values')
    sorted_indices_name = scope.get_unique_variable_name('sorted_indices')
    array_feat_extractor_output_name = scope.get_unique_variable_name(
        'array_feat_extractor_output')
    median_value_name = scope.get_unique_variable_name('median_value')
    comp_value_name = scope.get_unique_variable_name('comp_value')
    median_or_above_name = scope.get_unique_variable_name('median_or_above')
    median_idx_name = scope.get_unique_variable_name('median_idx')
    cast_result_name = scope.get_unique_variable_name('cast_result')
    reshaped_weights_name = scope.get_unique_variable_name('reshaped_weights')
    median_estimators_name = scope.get_unique_variable_name(
        'median_estimators')

    container.add_initializer(negate_name, proto_dtype,
                              [], [-1])
    container.add_initializer(estimators_weights_name,
                              proto_dtype,
                              [len(op.estimator_weights_)],
                              op.estimator_weights_)
    container.add_initializer(half_scalar_name, proto_dtype,
                              [], [0.5])
    container.add_initializer(last_index_name, onnx_proto.TensorProto.INT64,
                              [], [len(op.estimators_) - 1])

    concatenated_labels = _get_estimators_label(scope, operator,
                                                container, op)
    apply_mul(scope, [concatenated_labels, negate_name],
              negated_labels_name, container, broadcast=1)
    try:
        apply_topk(scope, negated_labels_name,
                   [sorted_values_name, sorted_indices_name],
                   container, k=len(op.estimators_))
    except TypeError:
        # onnxconverter-common < 1.7.0
        apply_topk(scope, [negated_labels_name],
                   [sorted_values_name, sorted_indices_name],
                   container, k=len(op.estimators_))
    container.add_node(
        'ArrayFeatureExtractor',
        [estimators_weights_name, sorted_indices_name],
        array_feat_extractor_output_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    apply_reshape(
        scope, array_feat_extractor_output_name, reshaped_weights_name,
        container, desired_shape=(-1, len(op.estimators_)))
    weights_cdf_name = cum_sum(
        scope, container, reshaped_weights_name,
        len(op.estimators_), proto_dtype)
    container.add_node(
        'ArrayFeatureExtractor', [weights_cdf_name, last_index_name],
        median_value_name, op_domain='ai.onnx.ml',
        name=scope.get_unique_operator_name('ArrayFeatureExtractor'))
    apply_mul(scope, [median_value_name, half_scalar_name],
              comp_value_name, container, broadcast=1)
    container.add_node(
        'Less', [weights_cdf_name, comp_value_name],
        median_or_above_name,
        name=scope.get_unique_operator_name('Less'))
    apply_cast(scope, median_or_above_name, cast_result_name,
               container, to=proto_dtype)
    container.add_node('ArgMin', cast_result_name,
                       median_idx_name,
                       name=scope.get_unique_operator_name('ArgMin'), axis=1)
    _apply_gather_elements(
        scope, container, [sorted_indices_name, median_idx_name],
        median_estimators_name, axis=1, dim=len(op.estimators_),
        zero_type=onnx_proto.TensorProto.INT64, suffix="A")
    output_name = operator.output_full_names[0]
    _apply_gather_elements(
        scope, container, [concatenated_labels, median_estimators_name],
        output_name, axis=1, dim=len(op.estimators_),
        zero_type=proto_dtype, suffix="B")


register_converter('SklearnAdaBoostClassifier',
                   convert_sklearn_ada_boost_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True],
                            'raw_scores': [True, False]})
register_converter('SklearnAdaBoostRegressor',
                   convert_sklearn_ada_boost_regressor)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from ..common._apply_operation import apply_cast
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_numpy_type)
from ..common._registration import register_converter
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs, get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs)
from ..proto import onnx_proto


def convert_sklearn_gradient_boosting_classifier(
        scope, operator, container, op_type='TreeEnsembleClassifier',
        op_domain='ai.onnx.ml', op_version=1):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    if op.loss not in ('deviance', 'log_loss'):
        raise NotImplementedError(
            "Loss '{0}' is not supported yet. You "
            "may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.".format(op.loss))

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs['name'] = scope.get_unique_operator_name(op_type)

    transform = 'LOGISTIC' if op.n_classes_ == 2 else 'SOFTMAX'
    if op.init == 'zero':
        loss = op._loss if hasattr(op, "_loss") else op.loss_
        base_values = np.zeros(loss.K)
    elif op.init is None:
        if hasattr(op.estimators_[0, 0], 'n_features_in_'):
            # sklearn >= 1.2
            n_features = op.estimators_[0, 0].n_features_in_
        else:
            # sklearn < 1.2
            n_features = op.estimators_[0, 0].n_features_
        x0 = np.zeros((1, n_features))
        if hasattr(op, '_raw_predict_init'):
            # sklearn >= 0.21
            base_values = op._raw_predict_init(x0).ravel()
        elif hasattr(op, '_init_decision_function'):
            # sklearn >= 0.20 and sklearn < 0.21
            base_values = op._init_decision_function(x0).ravel()
        else:
            raise RuntimeError("scikit-learn < 0.19 is not supported.")
    else:
        raise NotImplementedError(
            'Setting init to an estimator is not supported, you may raise an '
            'issue at https://github.com/onnx/sklearn-onnx/issues.')

    attrs['base_values'] = [float(v) for v in base_values]
    options = container.get_options(op, dict(raw_scores=False))
    if not options['raw_scores']:
        attrs['post_transform'] = transform

    classes = op.classes_
    if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
        class_labels = [int(i) for i in classes]
        attrs['classlabels_int64s'] = class_labels
    elif all(isinstance(i, str) for i in classes):
        class_labels = [str(i) for i in classes]
        attrs['classlabels_strings'] = class_labels
    else:
        raise ValueError('Labels must be all integer or all strings.')

    tree_weight = op.learning_rate
    n_est = (op.n_estimators_ if hasattr(op, 'n_estimators_') else
             op.n_estimators)
    if op.n_classes_ == 2:
        for tree_id in range(n_est):
            tree = op.estimators_[tree_id][0].tree_
            add_tree_to_attribute_pairs(attrs, True, tree, tree_id,
                                        tree_weight, 0, False, True,
                                        dtype=dtype)
    else:
        for i in range(n_est):
            for c in range(op.n_classes_):
                tree_id = i * op.n_classes_ + c
                tree = op.estimators_[i][c].tree_
                add_tree_to_attribute_pairs(attrs, True, tree, tree_id,
                                            tree_weight, c, False, True,
                                            dtype=dtype)

    if dtype is not None:
        for k in attrs:
            if k in ('nodes_values', 'class_weights',
                     'target_weights', 'nodes_hitrates',
                     'base_values'):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if isinstance(operator.inputs[0].type, BooleanTensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, input_name, cast_input_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        input_name = cast_input_name
    container.add_node(
        op_type, input_name,
        [operator.outputs[0].full_name, operator.outputs[1].full_name],
        op_domain=op_domain, op_version=op_version, **attrs)


def convert_sklearn_gradient_boosting_regressor(
        scope, operator, container, op_type='TreeEnsembleRegressor',
        op_domain='ai.onnx.ml', op_version=1):
    op = operator.raw_operator
    attrs = get_default_tree_regressor_attribute_pairs()
    attrs['name'] = scope.get_unique_operator_name(op_type)
    attrs['n_targets'] = 1

    if op.init == 'zero':
        loss = op._loss if hasattr(op, "_loss") else op.loss_
        cst = np.zeros(loss.K)
    elif op.init is None:
        # constant_ was introduced in scikit-learn 0.21.
        if hasattr(op.init_, 'constant_'):
            cst = [float(x) for x in op.init_.constant_]
        elif op.loss == 'ls':
            cst = [op.init_.mean]
        else:
            cst = [op.init_.quantile]
    else:
        raise NotImplementedError(
            'Setting init to an estimator is not supported, you may raise an '
            'issue at https://github.com/onnx/sklearn-onnx/issues.')

    attrs['base_values'] = [float(x) for x in cst]

    tree_weight = op.learning_rate
    n_est = (op.n_estimators_ if hasattr(op, 'n_estimators_') else
             op.n_estimators)
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    for i in range(n_est):
        tree = op.estimators_[i][0].tree_
        tree_id = i
        add_tree_to_attribute_pairs(attrs, False, tree, tree_id, tree_weight,
                                    0, False, True, dtype=dtype)

    if dtype is not None:
        for k in attrs:
            if k in ('nodes_values', 'class_weights',
                     'target_weights', 'nodes_hitrates',
                     'base_values'):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name('cast_input')

        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=onnx_proto.TensorProto.FLOAT)
        input_name = cast_input_name

    container.add_node(
        op_type, input_name, operator.output_full_names,
        op_domain=op_domain, op_version=op_version, **attrs)


register_converter('SklearnGradientBoostingClassifier',
                   convert_sklearn_gradient_boosting_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'raw_scores': [True, False],
                            'output_class_labels': [False, True],
                            'nocl': [True, False]})
register_converter('SklearnGradientBoostingRegressor',
                   convert_sklearn_gradient_boosting_regressor)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from scipy.linalg import solve_triangular
from sklearn.gaussian_process.kernels import ConstantKernel as C, RBF
try:
    from sklearn.gaussian_process._gpc import LAMBDAS, COEFS
except ImportError:
    LAMBDAS, COEFS = None, None
from ..proto import onnx_proto
from ..common.data_types import guess_numpy_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import (
    OnnxAdd, OnnxSqrt, OnnxMatMul, OnnxSub, OnnxReduceSumApi11,
    OnnxMul, OnnxMax, OnnxReshapeApi13, OnnxDiv, OnnxNot,
    OnnxReciprocal, OnnxCast, OnnxLess,
    OnnxPow, OnnxNeg, OnnxConcat, OnnxArrayFeatureExtractor,
    OnnxTranspose,
)
from ..algebra.custom_ops import OnnxSolve
try:
    from ..algebra.onnx_ops import OnnxConstantOfShape
except ImportError:
    OnnxConstantOfShape = None
try:
    from ..algebra.onnx_ops import OnnxErf
except ImportError:
    OnnxErf = None
try:
    from ..algebra.onnx_ops import OnnxEinsum
except ImportError:
    OnnxEinsum = None
from ._gp_kernels import (
    convert_kernel_diag,
    convert_kernel,
    _zero_vector_of_size
)


def convert_gaussian_process_regressor(scope: Scope, operator: Operator,
                                       container: ModelComponentContainer):
    """
    The method *predict* from class *GaussianProcessRegressor*
    may cache some results if it is called with parameter
    ``return_std=True`` or ``return_cov=True``. This converter
    needs to be called with theses options to enable
    the second results.
    See example :ref:`l-gpr-example` to see how to
    use this converter which does not behave exactly
    as the others.
    """
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset
    if opv is None:
        raise RuntimeError("container.target_opset must not be None")
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    options = container.get_options(
        op, dict(return_cov=False, return_std=False, optim=None))
    if hasattr(op, 'kernel_') and op.kernel_ is not None:
        kernel = op.kernel_
    elif op.kernel is None:
        kernel = (C(1.0, constant_value_bounds="fixed") *
                  RBF(1.0, length_scale_bounds="fixed"))
    else:
        kernel = op.kernel

    if not hasattr(op, "X_train_") or op.X_train_ is None:
        out0 = _zero_vector_of_size(X, keepdims=1, output_names=out[:1],
                                    dtype=dtype, op_version=opv)

        outputs = [out0]
        if options['return_cov']:
            outputs.append(convert_kernel(
                kernel, X, output_names=out[1:],
                dtype=dtype, op_version=opv))
        if options['return_std']:
            outputs.append(
                OnnxSqrt(
                    convert_kernel_diag(
                        kernel, X, dtype=dtype, op_version=opv),
                    output_names=out[1:], op_version=opv))
    else:
        # Code scikit-learn
        # K_trans = self.kernel_(X, self.X_train_)
        # y_mean = K_trans.dot(self.alpha_)  # Line 4 (y_mean = f_star)
        # y_mean = self._y_train_mean + y_mean * self._y_train_std

        k_trans = convert_kernel(
            kernel, X, x_train=op.X_train_.astype(dtype),
            dtype=dtype, optim=options.get('optim', None),
            op_version=opv)
        k_trans.set_onnx_name_prefix('kgpd')
        y_mean_b = OnnxMatMul(k_trans, op.alpha_.astype(dtype), op_version=opv)

        mean_y = op._y_train_mean.astype(dtype)
        if len(mean_y.shape) == 1:
            mean_y = mean_y.reshape(mean_y.shape + (1,))

        if not hasattr(op, '_y_train_std') or op._y_train_std == 1:
            if isinstance(y_mean_b, (np.float32, np.float64)):
                y_mean_b = np.array([y_mean_b])
            if isinstance(mean_y, (np.float32, np.float64)):
                mean_y = np.array([mean_y])
            y_mean = OnnxAdd(y_mean_b, mean_y, op_version=opv)
        else:
            # A bug was fixed in 0.23 and it changed
            # the predictions when return_std is True.
            # See https://github.com/scikit-learn/scikit-learn/pull/15782.
            # y_mean = self._y_train_std * y_mean + self._y_train_mean
            var_y = op._y_train_std.astype(dtype)
            if len(var_y.shape) == 1:
                var_y = var_y.reshape(var_y.shape + (1,))
            if isinstance(var_y, (np.float32, np.float64)):
                var_y = np.array([var_y])
            if isinstance(mean_y, (np.float32, np.float64)):
                mean_y = np.array([mean_y])
            y_mean = OnnxAdd(
                OnnxMul(y_mean_b, var_y, op_version=opv),
                mean_y, op_version=opv)

        y_mean.set_onnx_name_prefix('gpr')
        y_mean_reshaped = OnnxReshapeApi13(
            y_mean, np.array([-1, 1], dtype=np.int64),
            op_version=opv, output_names=out[:1])
        outputs = [y_mean_reshaped]

        if options['return_cov']:
            raise NotImplementedError()
        if options['return_std']:
            if hasattr(op, '_K_inv') and op._K_inv is not None:
                # scikit-learn < 0.24.2
                _K_inv = op._K_inv
            else:
                # scikit-learn >= 0.24.2
                L_inv = solve_triangular(op.L_.T, np.eye(op.L_.shape[0]))
                _K_inv = L_inv.dot(L_inv.T)

            # y_var = self.kernel_.diag(X)
            y_var = convert_kernel_diag(kernel, X, dtype=dtype,
                                        optim=options.get('optim', None),
                                        op_version=opv)

            # y_var -= np.einsum("ij,ij->i",
            #       np.dot(K_trans, self._K_inv), K_trans)
            k_dot = OnnxMatMul(k_trans, _K_inv.astype(dtype), op_version=opv)
            ys_var = OnnxSub(
                y_var, OnnxReduceSumApi11(
                    OnnxMul(k_dot, k_trans, op_version=opv),
                    axes=[1], keepdims=0, op_version=opv),
                op_version=opv)

            # y_var_negative = y_var < 0
            # if np.any(y_var_negative):
            #     y_var[y_var_negative] = 0.0
            ys0_var = OnnxMax(ys_var, np.array([0], dtype=dtype),
                              op_version=opv)

            if hasattr(op, '_y_train_std') and op._y_train_std != 1:
                # y_var = y_var * self._y_train_std**2
                ys0_var = OnnxMul(ys0_var, var_y ** 2, op_version=opv)

            # var = np.sqrt(ys0_var)
            var = OnnxSqrt(ys0_var, output_names=out[1:], op_version=opv)
            var.set_onnx_name_prefix('gprv')
            outputs.append(var)

    for o in outputs:
        o.add_to(scope, container)


def convert_gaussian_process_classifier(scope: Scope, operator: Operator,
                                        container: ModelComponentContainer):
    """
    The method *predict* from class *GaussianProcessClassifier*
    may cache some results if it is called with parameter
    ``return_std=True`` or ``return_cov=True``. This converter
    needs to be called with theses options to enable
    the second results.
    See example :ref:`l-gpr-example` to see how to
    use this converter which does not behave exactly
    as the others.
    """
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    op_est = operator.raw_operator.base_estimator_
    opv = container.target_opset
    if opv is None:
        raise RuntimeError("container.target_opset must not be None")
    if OnnxEinsum is None or OnnxErf is None:
        raise RuntimeError(
            "target opset must be >= 12 for operator 'einsum' and 'erf'.")
    if LAMBDAS is None:
        raise RuntimeError("Only scikit-learn>=0.22 is supported.")
    outputs = []

    options = container.get_options(op, dict(optim=None))
    if hasattr(op, 'kernel_') and op.kernel_ is not None:
        kernel = op.kernel_
    elif op.kernel is None:
        kernel = (C(1.0, constant_value_bounds="fixed") *
                  RBF(1.0, length_scale_bounds="fixed"))
    else:
        kernel = op.kernel

    if not hasattr(op_est, 'X_train_'):
        raise NotImplementedError("Only binary classification is iplemented.")
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32
    K_starT = convert_kernel(
        kernel, X, x_train=op_est.X_train_.astype(dtype), dtype=dtype,
        optim=options.get('optim', None), op_version=opv)
    K_star = OnnxTranspose(K_starT, op_version=opv)
    K_star.set_onnx_name_prefix('kstar')

    # common
    # f_star = K_star.T.dot(self.y_train_ - self.pi_)
    f_star_right = (op_est.y_train_ - op_est.pi_).astype(
        dtype).reshape((-1, 1))
    f_star = OnnxMatMul(K_starT, f_star_right, op_version=opv)
    f_star.set_onnx_name_prefix('f_star')

    best = OnnxCast(
        OnnxNot(
            OnnxLess(f_star, np.array([0], dtype=dtype), op_version=opv),
            op_version=opv),
        to=onnx_proto.TensorProto.INT64, op_version=opv)
    classes = OnnxArrayFeatureExtractor(op.classes_.astype(np.int64), best)
    labels = OnnxTranspose(classes, op_version=opv, output_names=out[:1])
    labels.set_onnx_name_prefix('labels')
    outputs.append(labels)

    # predict_proba
    # a x = b, x = a^-1 b
    # v = solve(self.L_, self.W_sr_[:, np.newaxis] * K_star)  # Line 5
    v = OnnxSolve(op_est.L_.astype(dtype),
                  OnnxMul(op_est.W_sr_[:, np.newaxis].astype(dtype),
                          K_star, op_version=opv),
                  op_version=opv)
    v.set_onnx_name_prefix('solve')

    # var_f_star = self.kernel_.diag(X) - np.einsum("ij,ij->j", v, v)
    var_f_star_kernel = convert_kernel_diag(
        kernel, X, dtype=dtype,
        optim=options.get('optim', None), op_version=opv)
    var_f_star_kernel.set_onnx_name_prefix('diag')
    var_f_star = OnnxSub(var_f_star_kernel,
                         OnnxEinsum(v, v, equation="ij,ij->j",
                                    op_version=opv),
                         op_version=opv)
    var_f_star.set_onnx_name_prefix('var_f_star')

    # alpha = 1 / (2 * var_f_star)
    alpha = OnnxReciprocal(OnnxMul(var_f_star, np.array([2], dtype=dtype),
                                   op_version=opv),
                           op_version=opv)
    alpha.set_onnx_name_prefix('alpha')

    # gamma = LAMBDAS * f_star
    gamma = OnnxMul(LAMBDAS.astype(dtype),
                    OnnxReshapeApi13(
                        f_star, np.array([1, -1], dtype=np.int64),
                        op_version=opv),
                    op_version=opv)
    gamma.set_onnx_name_prefix('gamma')

    # integrals = np.sqrt(np.pi / alpha) *
    #               erf(gamma * np.sqrt(alpha / (alpha + LAMBDAS**2))) /
    #               (2 * np.sqrt(var_f_star * 2 * np.pi))
    integrals_1 = OnnxSqrt(OnnxDiv(np.array([np.pi], dtype=dtype),
                                   alpha, op_version=opv),
                           op_version=opv)
    integrals_1.set_onnx_name_prefix('int1')

    integrals_2_1 = OnnxAdd(alpha, OnnxPow(LAMBDAS.astype(dtype),
                                           np.array([2], dtype=dtype),
                                           op_version=opv),
                            op_version=opv)
    integrals_2_1.set_onnx_name_prefix('int21')

    integrals_2_2 = OnnxSqrt(OnnxDiv(alpha, integrals_2_1, op_version=opv),
                             op_version=opv)
    integrals_2_2.set_onnx_name_prefix('int22')

    integrals_div = OnnxMul(
        np.array([2], dtype=dtype),
        OnnxSqrt(
            OnnxMul(
                OnnxMul(var_f_star, np.array([2], dtype=dtype),
                        op_version=opv),
                np.array([np.pi], dtype=dtype), op_version=opv),
            op_version=opv),
        op_version=opv)
    integrals_div.set_onnx_name_prefix('intdiv')

    integrals = OnnxMul(
        integrals_1,
        OnnxDiv(OnnxErf(OnnxMul(gamma, integrals_2_2, op_version=opv),
                        op_version=opv),
                integrals_div, op_version=opv),
        op_version=opv)
    integrals.set_onnx_name_prefix('integrals')

    # pi_star = (COEFS * integrals).sum(axis=0) + .5 * COEFS.sum()
    coef_sum = (.5 * COEFS.sum()).astype(dtype)
    if not isinstance(coef_sum, np.ndarray):
        coef_sum = np.array([coef_sum])
    pi_star = OnnxAdd(
        OnnxReduceSumApi11(
            OnnxMul(COEFS.astype(dtype), integrals, op_version=opv),
            op_version=opv, axes=[0]),
        coef_sum, op_version=opv)
    pi_star.set_onnx_name_prefix('pi_star')

    pi_star = OnnxReshapeApi13(
        pi_star, np.array([-1, 1], dtype=np.int64),
        op_version=opv)
    pi_star.set_onnx_name_prefix('pi_star2')
    final = OnnxConcat(
        OnnxAdd(OnnxNeg(pi_star, op_version=opv),
                np.array([1], dtype=dtype),
                op_version=opv),
        pi_star, op_version=opv, axis=1,
        output_names=out[1:2])
    outputs.append(final)

    for o in outputs:
        o.add_to(scope, container)


if OnnxConstantOfShape is not None:
    register_converter('SklearnGaussianProcessRegressor',
                       convert_gaussian_process_regressor,
                       options={'return_cov': [False, True],
                                'return_std': [False, True],
                                'optim': [None, 'cdist']})

if OnnxEinsum is not None and OnnxErf is not None:
    register_converter('SklearnGaussianProcessClassifier',
                       convert_gaussian_process_classifier,
                       options={'optim': [None, 'cdist'],
                                'nocl': [False, True],
                                'output_class_labels': [False, True],
                                'zipmap': [False, True]})



# SPDX-License-Identifier: Apache-2.0

import warnings
from typing import List
import numpy as np
from onnx.helper import (
    make_node, make_graph, make_model, make_tensor_value_info,
    TensorProto)
from onnx.numpy_helper import from_array
from onnx import onnx_pb as onnx_proto
from sklearn.base import BaseEstimator
from ..common.data_types import (
    Int64TensorType, FloatTensorType, DoubleTensorType,
    guess_numpy_type, guess_proto_type)
from ..common._topology import Scope, Operator, Variable, OPSET_TO_IR_VERSION
from ..common._container import ModelComponentContainer
from ..common.utils import (
    check_input_and_output_types,
    check_input_and_output_numbers,
    get_producer, get_producer_version,
    get_domain, get_model_version)
from .. import update_registered_converter
from .._supported_operators import _get_sklearn_operator_name
from ..algebra.onnx_ops import (
    OnnxIdentity, OnnxMatMul, OnnxGather, OnnxConcat, OnnxReshapeApi13,
    OnnxTreeEnsembleRegressor_1, OnnxOneHotEncoder, OnnxCast)
from .woe_transformer import WOETransformer


def woe_parser(scope: Scope, model: BaseEstimator,
               inputs: List[Variable], custom_parsers: dict = None):
    "ONNX parser for WOETransformer: defines the output type."
    alias = _get_sklearn_operator_name(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    output = scope.declare_local_variable(
        "encoding", inputs[0].type.__class__())
    this_operator.inputs = inputs
    this_operator.outputs.append(output)
    return this_operator.outputs


def woe_shape_calculator(operator: Operator):
    "ONNX shape calculator for WOETransformer: defines the output shape."
    type_list = [Int64TensorType, FloatTensorType, DoubleTensorType]
    check_input_and_output_types(
        operator, good_input_types=type_list, good_output_types=type_list)
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=1)
    op = operator.raw_operator
    x = operator.inputs[0]
    N = x.get_first_dimension()
    if op.onehot:
        C = 0
        for ext in op.intervals_:
            if ext is None:
                C += 1
            else:
                C += len(ext)
        operator.outputs[0].type.shape = [N, C]
    else:
        C = len(op.intervals_)
        operator.outputs[0].type.shape = [N, C]


class Tree:

    class Node:

        def __init__(self, parent, is_left, is_leaf, feature,
                     threshold, value, index=-1):
            self.parent = parent
            self.is_left = is_left
            self.is_leaf = is_leaf
            self.feature = feature
            self.threshold = threshold
            self.value = value
            self.index = index

        def __str__(self):
            return (
                "Node(%s, is_left=%r, is_leaf=%r, feature=%r, "
                "threshold=%r, value=%r, index=%r)%s" % (
                    self.parent if isinstance(self.parent, int)
                    else "id%r" % id(self.parent),
                    self.is_left, self.is_leaf, self.feature,
                    self.threshold, self.value, self.index,
                    "  # %s %r -> %r%s%s%s" % (
                        self.onnx_mode, self.onnx_threshold, self.onnx_value,
                        " -- %r" % self.intervals_
                        if hasattr(self, 'intervals_') else '',
                        " LL %r" % self.intervals_left_
                        if hasattr(self, 'intervals_left_') else '',
                        " RR %r" % self.intervals_right_
                        if hasattr(self, 'intervals_right_') else '')))

        @property
        def onnx_value(self):
            return self.value if self.value is not None else 0

        @property
        def onnx_threshold(self):
            if self.is_leaf:
                return self.threshold
            return self.threshold[0]

        @property
        def onnx_mode(self):
            # 'BRANCH_LEQ', 'BRANCH_LT', 'BRANCH_GTE', 'BRANCH_GT',
            # 'BRANCH_EQ', 'BRANCH_NEQ', 'LEAF'
            if self.is_leaf:
                return 'LEAF'
            if self.threshold[1]:
                return 'BRANCH_LEQ'
            return 'BRANCH_LT'

        @staticmethod
        def _is_on_left_side(th, kind, x, leq, left_right_extremity):
            if kind not in ('BRANCH_LEQ', 'BRANCH_LT'):
                raise NotImplementedError(
                    "Not implemented for mode %r." % kind)
            if x < th:
                return False
            if x > th:
                return True
            if not left_right_extremity and not leq:
                return True
            if left_right_extremity and not leq:
                return False
            if kind == 'BRANCH_LEQ' and leq:
                return False
            if kind == 'BRANCH_LT' and not leq:
                return False
            return True

        def is_on_left_side(self, x, leq, left_right_extremity):
            th = self.threshold[0]
            kind = self.onnx_mode
            res = Tree.Node._is_on_left_side(
                th, kind, x, leq, left_right_extremity)
            return res

    def __init__(self):
        self.nodes = []
        self.leave_values = set()

    def __str__(self):
        mapping = {}
        for n in self.nodes:
            mapping[str(id(n))] = str(len(mapping))
        rows = []
        for n in self.nodes:
            rows.append("id%r = %s" % (id(n), n))
        res = "\n".join(rows)
        for k, v in mapping.items():
            res = res.replace("id" + k, "n" + v)
        return res

    def add_node(self, parent, is_left, is_leaf, feature, threshold,
                 value=None, index=-1):
        if is_leaf and value is None:
            raise ValueError("value must be specified when is_leaf=True.")
        if not is_leaf and value is not None:
            raise ValueError("value must not be specified when is_leaf=False.")
        node = Tree.Node(parent, is_left, is_leaf, feature, threshold,
                         value, index=index)
        self.nodes.append(node)
        if is_leaf:
            if value in self.leave_values:
                raise RuntimeError(
                    "The tree must contain unique tree value, %r "
                    "already in %r.\n%s" % (
                        value, self.leave_values, str(self)))
            self.leave_values.add(value)
        return node

    def _update_feature(self, feature):
        "Change the feature index."
        for node in self.nodes:
            node.feature = feature

    def onnx_attributes(self):
        """
        See `TreeEnsembleRegressor
        <https://github.com/onnx/onnx/blob/master/docs/
        Operators-ml.md#ai.onnx.ml.TreeEnsembleRegressor>`_.
        """
        atts = dict(
            aggregate_function='SUM',
            base_values=[float(0)],
            n_targets=1,
            nodes_featureids=[n.feature for n in self.nodes],
            nodes_missing_value_tracks_true=[0 for n in self.nodes],
            nodes_modes=[n.onnx_mode for n in self.nodes],
            nodes_nodeids=[i for i in range(len(self.nodes))],
            nodes_treeids=[0 for n in self.nodes],
            nodes_values=[float(n.onnx_threshold) for n in self.nodes],
            post_transform='NONE',
            target_ids=[0 for n in self.nodes if n.onnx_mode == 'LEAF'],
            target_nodeids=[i for i, n in enumerate(self.nodes)
                            if n.onnx_mode == 'LEAF'],
            target_treeids=[0 for n in self.nodes
                            if n.onnx_mode == 'LEAF'],
            target_weights=[float(n.onnx_value) for n in self.nodes
                            if n.onnx_mode == 'LEAF'])

        ids = {id(n): (i, n) for i, n in enumerate(self.nodes)}
        nodes_truenodeids = [0 for n in self.nodes]  # right
        nodes_falsenodeids = [0 for n in self.nodes]  # left
        for i, n in enumerate(self.nodes):
            if n.parent == -1:
                continue
            idp = id(n.parent)
            val = ids[idp]
            if n.is_left:
                nodes_truenodeids[val[0]] = i
            else:
                nodes_falsenodeids[val[0]] = i
        atts.update(dict(
            nodes_falsenodeids=nodes_falsenodeids,
            nodes_truenodeids=nodes_truenodeids))
        if len(atts['target_weights']) != len(set(atts['target_weights'])):
            warnings.warn(
                "All targets should be unique %r." % atts['target_weights'])
        return atts

    def mapping(self, intervals):
        """
        Maps every leaf target to the list of intervals
        it intersects. It creates attributes `intervals_`, `intervals_left_`,
        `intervals_rights_` as dictionary `{idx: interval}`
        each side intersects.
        """
        def process(node, intervals):
            if hasattr(node, 'intervals_'):
                return 0

            if node.parent is None or node.parent == -1:
                node.intervals_ = intervals
            else:
                if not hasattr(node.parent, 'intervals_'):
                    return 0
                node.intervals_ = (
                    node.parent.intervals_left_ if node.is_left
                    else node.parent.intervals_right_)

            if node.value is not None:
                # leaf
                return 1

            left = {}
            right = {}
            for k, v in node.intervals_.items():
                deca = node.is_on_left_side(v[0], v[2], False)
                decb = node.is_on_left_side(v[1], v[3], True)
                if not decb:
                    assert not deca
                    left[k] = v
                elif deca:
                    assert decb
                    right[k] = v
                elif decb and not deca:
                    left[k] = v
                    right[k] = v

            node.intervals_left_ = left
            node.intervals_right_ = right
            return 1

        for node in self.nodes:
            for at in ['intervals_', 'intervals_left_', 'intervals_right_']:
                if hasattr(node, at):
                    delattr(node, at)

        d_intervals = {i: t for i, t in enumerate(intervals)}
        changes = 1
        while changes > 0:
            changes = 0
            for node in self.nodes:
                changes += process(node, d_intervals)

        # final
        mapping = {}
        for node in self.nodes:
            if node.value is None:
                # not a leaf
                continue
            mapping[node.onnx_value] = node.intervals_
        return mapping


def digitize2tree(bins, right=False, feature=0):
    ascending = len(bins) <= 1 or bins[0] < bins[1]

    if not ascending:
        bins2 = bins[::-1]
        cl = digitize2tree(bins2, right=right, feature=feature)
        n = len(bins)
        for i in range(cl.tree_.value.shape[0]):
            cl.tree_.value[i, 0, 0] = n - cl.tree_.value[i, 0, 0]
        return cl

    tree = Tree()
    values = []
    UNUSED = np.nan
    n_nodes = []

    def add_root(index):
        if index < 0 or index >= len(bins):
            raise IndexError(  # pragma: no cover
                "Unexpected index %d / len(bins)=%d." % (
                    index, len(bins)))
        parent = -1
        is_left = False
        is_leaf = False
        threshold = bins[index]
        n = tree.add_node(parent, is_left, is_leaf, 0, threshold)
        values.append(UNUSED)
        n_nodes.append(n)
        return n

    def add_nodes(parent, i, j, is_left):
        # add for bins[i:j] (j excluded)
        if is_left:
            # it means j is the parent split
            if i == j:
                # leaf
                value = parent.index * 2
                n = tree.add_node(
                    parent, is_left, True, 0, 0, value=value, index=i)
                n_nodes.append(n)
                values.append(i)
                return n
            if i + 1 == j:
                # split
                values.append(UNUSED)
                th = bins[i]
                n = tree.add_node(parent, is_left, False, 0, th, index=i)
                n_nodes.append(n)
                add_nodes(n, i, i, True)
                add_nodes(n, i, j, False)
                return n
            if i < j:
                # split
                values.append(UNUSED)
                index = (i + j) // 2
                th = bins[index]
                n = tree.add_node(parent, is_left, False, 0, th, index=index)
                n_nodes.append(n)
                add_nodes(n, i, index, True)
                add_nodes(n, index, j, False)
                return n
        else:
            # it means i is the parent split
            if i + 1 == j:
                # leaf
                value = parent.index * 2 + 1
                values.append(j)
                n = tree.add_node(
                    parent, is_left, True, 0, 0, value=value, index=j)
                n_nodes.append(n)
                return n
            if i + 1 < j:
                # split
                values.append(UNUSED)
                index = (i + j) // 2
                th = bins[index]
                n = tree.add_node(parent, is_left, False, 0, th, index=index)
                n_nodes.append(n)
                add_nodes(n, i, index, True)
                add_nodes(n, index, j, False)
                return n
        raise NotImplementedError(  # pragma: no cover
            "Unexpected case where i=%r, j=%r, is_left=%r." % (
                i, j, is_left))

    index = len(bins) // 2
    root = add_root(index)
    add_nodes(root, 0, index, True)
    add_nodes(root, index, len(bins), False)
    tree._update_feature(feature)
    return tree


def _mapping2matrix(mapping, value_mapping, weights, dtype):
    rev = {v: k for k, v in enumerate(value_mapping)}
    rows = int(max(rev[k] for k in mapping)) + 1
    cols = max(max(v) if v else 0 for v in mapping.values()) + 1
    mat = np.zeros((rows, cols), dtype=dtype)
    for k, intervals in mapping.items():
        for idx in intervals:
            mat[rev[k], idx] = weights[idx]

    # Remove all empty columns.
    total = mat.sum(axis=0, keepdims=0)
    return mat[:, total > 0].copy()


def _mapping_to_key_value(mapping, weights):
    key_value = {}
    for k, v in sorted(mapping.items()):
        if len(v) == 0:
            continue
        if len(v) != 1:
            raise RuntimeError(
                'Intervals overlops in mapping %r.' % mapping)
        value = list(v)[0]
        key_value[float(k)] = float(weights[value])
    return key_value


def woe_converter(scope: Scope, operator: Operator,
                  container: ModelComponentContainer):
    """
    ONNX Converter for WOETransformer.
    It follows *skl2onnx* API.
    The logic behind the converter is summarized
    by the following picture:

    .. image:: images/woe.png
    """
    op = operator.raw_operator
    X = operator.inputs[0]
    output = operator.outputs[0]
    opv = container.target_opset
    new_shape = np.array([-1, 1], dtype=np.int64)
    vector_shape = np.array([-1], dtype=np.int64)
    dtype = guess_numpy_type(X.type)
    proto_type = guess_proto_type(X.type)
    verbose = getattr(container, 'verbose', 0)

    columns = []

    thresholds = op._decision_thresholds(add_index=False)
    conc = None
    for i, threshold in enumerate(thresholds):
        if threshold is None:
            # Passthrough columns
            index = np.array([i], dtype=np.int64)
            columns.append(
                OnnxReshapeApi13(
                    OnnxGather(X, index, op_version=opv, axis=1),
                    new_shape, op_version=opv))
            continue

        # encoding columns
        tree = digitize2tree(threshold, feature=i)

        atts = tree.onnx_attributes()
        mapping = tree.mapping(op.intervals_[i])

        if op.onehot:
            node = OnnxTreeEnsembleRegressor_1(
                X, op_version=1, domain='ai.onnx.ml', **atts)
            cats = list(sorted(set(int(n.onnx_value)
                                   for n in tree.nodes if n.is_leaf)))
            mat_mapping = _mapping2matrix(mapping, cats, op.weights_[i], dtype)
            if verbose > 1:
                print("[woe_converter] mapping=%r" % mapping)
            ohe = OnnxOneHotEncoder(
                OnnxReshapeApi13(
                    node, vector_shape, op_version=opv),
                op_version=opv, cats_int64s=cats)
            ren = OnnxMatMul(
                OnnxCast(ohe, op_version=opv, to=proto_type),
                mat_mapping, op_version=opv)
            columns.append(ren)
        else:
            key_value = _mapping_to_key_value(mapping, op.weights_[i])
            atts['target_weights'] = [
                key_value.get(v, 0.) for v in atts['target_weights']]
            if verbose > 1:
                print("[woe_converter] mapping=%r" % mapping)
                print("[woe_converter] key_value=%r" % key_value)
            node = OnnxTreeEnsembleRegressor_1(
                X, op_version=1, domain='ai.onnx.ml', **atts)
            lab = OnnxReshapeApi13(node, new_shape, op_version=opv)
            columns.append(lab)

    conc = OnnxConcat(*columns, op_version=opv, axis=1)
    final = OnnxIdentity(conc, output_names=[output], op_version=opv)
    final.add_to(scope, container)


def woe_transformer_to_onnx(op, opset=None):
    """
    ONNX Converter for WOETransformer.
    It uses ONNX API.
    The logic behind the converter is summarized
    by the following picture:

    .. image:: images/woe.png

    The converter only adds *opset* in the ONNX graph.
    It does not change the conversion depending on the opset value.
    """
    if op.onehot:
        C = 0
        for ext in op.intervals_:
            if ext is None:
                C += 1
            else:
                C += len(ext)
    else:
        C = len(op.intervals_)

    # inputs
    X = make_tensor_value_info(
        'X', TensorProto.FLOAT, [None, len(op.intervals_)])
    Y = make_tensor_value_info(
        'Y', TensorProto.FLOAT, [None, C])

    # nodes
    nodes = []
    columns = []
    inits = [from_array(np.array([-1, 1], dtype=np.int64), name='new_shape'),
             from_array(np.array([-1], dtype=np.int64), name='vector_shape')]
    thresholds = op._decision_thresholds(add_index=False)

    for i, threshold in enumerate(thresholds):
        if threshold is None:
            # Passthrough columns
            inits.append(from_array(
                np.array([i], dtype=np.int64), name='index%d' % i))
            nodes.append(make_node(
                'Gather', ['X', 'index%d' % i], ['col%d' % i], axis=1))
            nodes.append(make_node(
                'Reshape', ['col%d' % i, 'new_shape'], ['reshr%d' % i]))
            columns.append('reshr%d' % i)
            continue

        # encoding columns
        tree = digitize2tree(threshold, feature=i)
        mapping = tree.mapping(op.intervals_[i])

        atts = tree.onnx_attributes()

        if op.onehot:
            nodes.append(make_node(
                'TreeEnsembleRegressor', ['X'], ['rf%d' % i],
                domain='ai.onnx.ml', **atts))
            cats = list(sorted(set(int(n.onnx_value)
                                   for n in tree.nodes if n.is_leaf)))
            mat_mapping = _mapping2matrix(
                mapping, cats, op.weights_[i], np.float32)
            nodes.append(make_node(
                'Reshape', ['rf%d' % i, 'vector_shape'], ['resh%d' % i]))
            nodes.append(make_node(
                'OneHotEncoder', ['resh%d' % i], ['ohe%d' % i],
                domain='ai.onnx.ml', cats_int64s=cats))
            nodes.append(make_node(
                'Cast', ['ohe%d' % i], ['cast%d' % i], to=TensorProto.FLOAT))
            inits.append(from_array(mat_mapping, 'mat_map%i' % i))
            nodes.append(make_node(
                'MatMul', ['cast%d' % i, 'mat_map%i' % i], ["mul%d" % i]))
            columns.append("mul%d" % i)
        else:
            key_value = _mapping_to_key_value(mapping, op.weights_[i])
            atts['target_weights'] = [
                key_value.get(v, 0.) for v in atts['target_weights']]
            nodes.append(make_node(
                'TreeEnsembleRegressor', ['X'], ['rf%d' % i],
                domain='ai.onnx.ml', **atts))
            nodes.append(make_node(
                'Reshape', ['rf%d' % i, 'new_shape'], ['lab%d' % i]))
            columns.append("lab%d" % i)

    nodes.append(make_node(
        'Concat', columns, ['Y'], axis=1))

    # final graph
    graph_def = make_graph(nodes, 't1', [X], [Y], inits)
    model_def = make_model(graph_def, producer_name='skl2onnx')

    if opset is not None:
        del model_def.opset_import[:]
        op_set = model_def.opset_import.add()
        op_set.domain = ''
        op_set.version = opset
        op_set = model_def.opset_import.add()
        op_set.domain = 'ai.onnx.ml'
        op_set.version = 2
        irv = OPSET_TO_IR_VERSION.get(opset, onnx_proto.IR_VERSION)
        model_def.ir_version = irv

    model_def.producer_name = get_producer()
    model_def.producer_version = get_producer_version()
    model_def.domain = get_domain()
    model_def.model_version = get_model_version()
    model_def.doc_string = "WOETransformer"
    return model_def


def register():
    "Register converter for WOETransformer."
    update_registered_converter(
        WOETransformer, "Skl2onnxWOETransformer",
        woe_shape_calculator, woe_converter, parser=woe_parser)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import (
    apply_add, apply_argmax, apply_cast, apply_concat, apply_div, apply_exp,
    apply_log, apply_matmul, apply_mul, apply_pow,
    apply_reducesum, apply_reshape, apply_sub, apply_transpose)
from ..common.data_types import (
    BooleanTensorType, Int64TensorType, guess_proto_type)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_quadratic_discriminant_analysis_classifier(
        scope: Scope, operator: Operator, container: ModelComponentContainer):

    input_name = operator.inputs[0].full_name
    model = operator.raw_operator

    n_classes = len(model.classes_)

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if isinstance(operator.inputs[0].type,
                  (BooleanTensorType, Int64TensorType)):
        cast_input_name = scope.get_unique_variable_name('cast_input')
        apply_cast(scope, operator.input_full_names, cast_input_name,
                   container, to=proto_dtype)
        input_name = cast_input_name

    norm_array_name = []
    sum_array_name = []

    const_n05 = scope.get_unique_variable_name('const_n05')
    const_p2 = scope.get_unique_variable_name('const_p2')
    container.add_initializer(const_n05, proto_dtype, [], [-0.5])
    container.add_initializer(const_p2, proto_dtype, [], [2])

    for i in range(n_classes):
        R = model.rotations_[i]
        rotation_name = scope.get_unique_variable_name('rotations')
        container.add_initializer(rotation_name, proto_dtype,
                                  [R.shape[0], R.shape[1]], R)

        S = model.scalings_[i]
        scaling_name = scope.get_unique_variable_name('scalings')
        container.add_initializer(
            scaling_name, proto_dtype, [S.shape[0], ], S)

        mean = model.means_[i]
        mean_name = scope.get_unique_variable_name('means')
        container.add_initializer(mean_name, proto_dtype, mean.shape, mean)

        Xm_name = scope.get_unique_variable_name('Xm')
        apply_sub(scope, [input_name, mean_name], [Xm_name], container)

        s_pow_name = scope.get_unique_variable_name('s_pow_n05')
        apply_pow(scope, [scaling_name, const_n05], [s_pow_name], container)

        mul_name = scope.get_unique_variable_name('mul')
        apply_mul(scope, [rotation_name, s_pow_name], [mul_name], container)

        x2_name = scope.get_unique_variable_name('matmul')
        apply_matmul(scope, [Xm_name, mul_name], [x2_name], container)

        pow_x2_name = scope.get_unique_variable_name('pow_x2')
        apply_pow(scope, [x2_name, const_p2], [pow_x2_name], container)

        sum_name = scope.get_unique_variable_name('sum')
        apply_reducesum(scope, [pow_x2_name], [sum_name],
                        container, axes=[1], keepdims=1)
        norm_array_name.append(sum_name)

        log_name = scope.get_unique_variable_name('log')
        apply_log(scope, [scaling_name], [log_name], container)

        sum_log_name = scope.get_unique_variable_name('sum_log')
        apply_reducesum(
            scope, [log_name], [sum_log_name], container, keepdims=1)
        sum_array_name.append(sum_log_name)

    concat_norm_name = scope.get_unique_variable_name('concat_norm')
    apply_concat(scope, norm_array_name, [concat_norm_name], container)

    reshape_norm_name = scope.get_unique_variable_name('reshape_concat_norm')
    apply_reshape(scope, [concat_norm_name], [reshape_norm_name],
                  container, desired_shape=[n_classes, -1])

    transpose_norm_name = scope.get_unique_variable_name('transpose_norm')
    apply_transpose(scope, [reshape_norm_name], [transpose_norm_name],
                    container, perm=(1, 0))

    concat_logsum = scope.get_unique_variable_name('concat_logsum')
    apply_concat(scope, sum_array_name, [concat_logsum], container)

    add_norm2_u_name = scope.get_unique_variable_name('add_norm2_u')
    apply_add(scope, [transpose_norm_name, concat_logsum],
              [add_norm2_u_name], container)

    norm2_u_n05_name = scope.get_unique_variable_name('norm2_u_n05')
    apply_mul(
        scope, [const_n05, add_norm2_u_name], [norm2_u_n05_name], container)

    priors = scope.get_unique_variable_name('priors')
    container.add_initializer(
        priors, proto_dtype, [n_classes, ], model.priors_)
    log_p = scope.get_unique_variable_name('log_p')
    apply_log(scope, [priors], [log_p], container)

    decision_fun = scope.get_unique_variable_name('decision_fun')
    apply_add(scope, [norm2_u_n05_name, log_p], [decision_fun], container)

    argmax_out = scope.get_unique_variable_name('argmax_out')
    apply_argmax(scope, [decision_fun], [argmax_out], container, axis=1)

    classes = scope.get_unique_variable_name('classes')
    container.add_initializer(
        classes, onnx_proto.TensorProto.INT64, [n_classes], model.classes_)

    container.add_node(
        'ArrayFeatureExtractor',
        [classes, argmax_out],
        [operator.outputs[0].full_name],
        op_domain='ai.onnx.ml'
    )

    df_max = scope.get_unique_variable_name('df_max')
    if container.target_opset >= 18:
        axis_name = scope.get_unique_variable_name('axis')
        container.add_initializer(
            axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            'ReduceMax', [decision_fun, axis_name], [df_max])
    else:
        container.add_node(
            'ReduceMax', [decision_fun], [df_max], axes=[1])
    df_sub_max = scope.get_unique_variable_name('df_sub_max')
    apply_sub(scope, [decision_fun, df_max], [df_sub_max], container)
    likelihood = scope.get_unique_variable_name('likelihood')
    apply_exp(scope, [df_sub_max], [likelihood], container)
    likelihood_sum = scope.get_unique_variable_name('likelihood_sum')
    apply_reducesum(scope, [likelihood], [likelihood_sum], container,
                    axes=[1], keepdims=1)
    apply_div(scope, [likelihood, likelihood_sum],
              [operator.outputs[1].full_name], container, )


register_converter('SklearnQuadraticDiscriminantAnalysis',
                   convert_quadratic_discriminant_analysis_classifier,
                   options={'zipmap': [True, False, 'columns'],
                            'nocl': [True, False],
                            'output_class_labels': [False, True]})



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_tfidf_transformer_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1,
                                   output_count_range=1)
    C = operator.inputs[0].type.shape[1]
    operator.outputs[0].type.shape = [1, C]


register_shape_calculator('SklearnTfidfTransformer',
                          calculate_sklearn_tfidf_transformer_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import warnings
from uuid import uuid4
from .proto import get_latest_tested_opset_version
from .common._topology import convert_topology
from .common.utils_sklearn import _process_options
from ._parse import parse_sklearn_model

# Invoke the registration of all our converters and shape calculators.
from . import shape_calculators  # noqa
from . import operator_converters  # noqa


def convert_sklearn(model, name=None, initial_types=None, doc_string='',
                    target_opset=None, custom_conversion_functions=None,
                    custom_shape_calculators=None,
                    custom_parsers=None, options=None,
                    intermediate=False,
                    white_op=None, black_op=None, final_types=None,
                    dtype=None, naming=None, model_optim=True,
                    verbose=0):
    """
    This function produces an equivalent
    ONNX model of the given scikit-learn model.
    The supported converters is returned by function
    :func:`supported_converters <skl2onnx.supported_converters>`.

    For pipeline conversion, user needs to make sure each component
    is one of our supported items.
    This function converts the specified *scikit-learn* model
    into its *ONNX* counterpart.
    Note that for all conversions, initial types are required.
    *ONNX* model name can also be specified.

    :param model: A scikit-learn model
    :param initial_types: a python list.
        Each element is a tuple of a variable name
        and a type defined in `data_types.py`
    :param name: The name of the graph (type: GraphProto)
        in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for
        ONNX 1.2, and 8 for ONNX 1.3,
        if value is not specified, the function will
        choose the latest tested opset
        (see :py:func:`skl2onnx.get_latest_tested_opset_version`)
    :param custom_conversion_functions: a dictionary for
        specifying the user customized conversion function,
        it takes precedence over registered converters
    :param custom_shape_calculators: a dictionary for
        specifying the user customized shape calculator
        it takes precedence over registered shape calculators.
    :param custom_parsers: parsers determines which outputs
        is expected for which particular task,
        default parsers are defined for classifiers,
        regressors, pipeline but they can be rewritten,
        *custom_parsers* is a dictionary
        ``{ type: fct_parser(scope, model, inputs, custom_parsers=None) }``
    :param options: specific options given to converters
        (see :ref:`l-conv-options`)
    :param intermediate: if True, the function returns the
        converted model and the instance of :class:`Topology` used,
        it returns the converted model otherwise
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline,
        if empty, all are allowed
    :param black_op: black list of ONNX nodes
        allowed while converting a pipeline,
        if empty, none are blacklisted
    :param final_types: a python list. Works the same way as initial_types
        but not mandatory, it is used to overwrites the type
        (if type is not None) and the name of every output.
    :param dtype: removed in version 1.7.5, dtype is
        now inferred from input types,
        converters may add operators Cast to switch
        to double when it is necessary
    :param naming: the user may want to change the way intermediate
        are named, this parameter can be a string (a prefix) or a
        function, which signature is the following:
        `get_name(name, existing_names)`, the library will then
        check this name is unique and modify it if not
    :param model_optim: enable or disable model optimisation
        after the model was converted into onnx, it reduces the number
        of identity nodes
    :param verbose: display progress while converting a model
    :return: An ONNX model (type: ModelProto) which is
        equivalent to the input scikit-learn model

    Example of *initial_types*:
    Assume that the specified *scikit-learn* model takes
    a heterogeneous list as its input.
    If the first 5 elements are floats and the last 10 elements are integers,
    we need to specify initial types as below. The [None] in
    [None, 5] indicates the batch size here is unknown.

    ::

        from skl2onnx.common.data_types import FloatTensorType, Int64TensorType
        initial_type = [('float_input', FloatTensorType([None, 5])),
                        ('int64_input', Int64TensorType([None, 10]))]

    .. note::

        If a pipeline includes an instance of
        `ColumnTransformer <https://scikit-learn.org/stable/modules/
        generated/sklearn.compose.ColumnTransformer.html>`_,
        *scikit-learn* allow the user to specify columns by names.
        This option is not supported
        by *sklearn-onnx* as features names could be different
        in input data and the ONNX graph
        (defined by parameter *initial_types*), only integers are supported.

    .. _l-conv-options:

    Converters options
    ++++++++++++++++++

    Some ONNX operators exposes parameters *sklearn-onnx* cannot
    guess from the raw model. Some default values are usually suggested
    but the users may have to manually overwrite them. This need
    is not obvious to do when a model is included in a pipeline.
    That's why these options can be given to function *convert_sklearn*
    as a dictionary ``{model_type: parameters in a dictionary}`` or
    ``{model_id: parameters in a dictionary}``.
    Option *sep* is used to specify the delimiters between two words
    when the ONNX graph needs to tokenize a string.
    The default value is short and may not include all
    the necessary values. It can be overwritten as:

    ::

        extra = {TfidfVectorizer: {"separators": [' ', '[.]', '\\\\?',
                    ',', ';', ':', '\\\\!', '\\\\(', '\\\\)']}}
        model_onnx = convert_sklearn(
            model, "tfidf",
            initial_types=[("input", StringTensorType([None, 1]))],
            options=extra)

    But if a pipeline contains two model of the same class,
    it is possible to distinguish between the two with function *id*:

    ::

        extra = {id(model): {"separators": [' ', '.', '\\\\?', ',', ';',
                    ':', '\\\\!', '\\\\(', '\\\\)']}}
        model_onnx = convert_sklearn(
            pipeline, "pipeline-with-2-tfidf",
            initial_types=[("input", StringTensorType([None, 1]))],
            options=extra)

    It is used in example :ref:`l-example-tfidfvectorizer`.

    .. versionchanged:: 1.10.0
        Parameter *naming* was added.
    """
    if initial_types is None:
        if hasattr(model, 'infer_initial_types'):
            initial_types = model.infer_initial_types()
        else:
            raise ValueError('Initial types are required. See usage of '
                             'convert(...) in skl2onnx.convert for details')

    if name is None:
        name = str(uuid4().hex)
    if dtype is not None:
        warnings.warn(
            "Parameter dtype is no longer supported. "
            "It will be removed in 1.9.0.",
            DeprecationWarning)

    target_opset = (target_opset
                    if target_opset else get_latest_tested_opset_version())
    # Parse scikit-learn model as our internal data structure
    # (i.e., Topology)
    if verbose >= 1:
        print("[convert_sklearn] parse_sklearn_model")
    topology = parse_sklearn_model(
        model, initial_types, target_opset, custom_conversion_functions,
        custom_shape_calculators, custom_parsers, options=options,
        white_op=white_op, black_op=black_op,
        final_types=final_types, naming=naming)

    # Convert our Topology object into ONNX. The outcome is an ONNX model.
    options = _process_options(model, options)
    if verbose >= 1:
        print("[convert_sklearn] convert_topology")
    onnx_model = convert_topology(
        topology, name, doc_string, target_opset, options=options,
        remove_identity=model_optim and not intermediate, verbose=verbose)
    if verbose >= 1:
        print("[convert_sklearn] end")
        if verbose >= 2:
            scope = topology.scopes[0]
            print("---INPUTS---")
            for inp in scope.input_variables:
                print("  %r" % inp)
            print("---OUTPUTS---")
            for inp in scope.output_variables:
                print("  %r" % inp)
            print("---VARIABLES---")
            for k, v in sorted(scope.variables.items()):
                print("  %r: is.fed=%r is_leaf=%r - %r" % (
                    k, v.is_fed, v.is_leaf, v))
            print("---OPERATORS---")
            for k, v in sorted(scope.operators.items()):
                print("  %r: is.evaluated=%r - %r" % (
                    k, v.is_evaluated, v))

    return (onnx_model, topology) if intermediate else onnx_model


def to_onnx(model, X=None, name=None, initial_types=None,
            target_opset=None, options=None,
            white_op=None, black_op=None, final_types=None,
            dtype=None, naming=None, model_optim=True, verbose=0):
    """
    Calls :func:`convert_sklearn` with simplified parameters.

    :param model: model to convert
    :param X: training set, can be None, it is used to infered the
        input types (*initial_types*)
    :param initial_types: if X is None, then *initial_types* must be
        defined
    :param target_opset: conversion with a specific target opset
    :param options: specific options given to converters
        (see :ref:`l-conv-options`)
    :param name: name of the model
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline, if empty, all are allowed
    :param black_op: black list of ONNX nodes allowed
        while converting a pipeline, if empty, none are blacklisted
    :param final_types: a python list. Works the same way as initial_types
        but not mandatory, it is used to overwrites the type
        (if type is not None) and the name of every output.
    :param dtype: removed in version 1.7.5, dtype is now inferred from
        input types, converters may add operators Cast to switch to
        double when it is necessary
    :param naming: the user may want to change the way intermediate
        are named, this parameter can be a string (a prefix) or a
        function, which signature is the following:
        `get_name(name, existing_names)`, the library will then
        check this name is unique and modify it if not
    :param model_optim: enable or disable model optimisation
        after the model was converted into onnx, it reduces the number
        of identity nodes
    :param verbose: display progress while converting a model
    :return: converted model

    This function checks if the model inherits from class
    :class:`OnnxOperatorMixin`, it calls method *to_onnx*
    in that case otherwise it calls :func:`convert_sklearn`.

    .. versionchanged:: 1.10.0
        Parameter *naming* was added.
    """
    from .algebra.onnx_operator_mixin import OnnxOperatorMixin
    from .algebra.type_helper import guess_initial_types

    if isinstance(model, OnnxOperatorMixin):
        if options is not None:
            raise NotImplementedError(
                "options not yet implemented for OnnxOperatorMixin.")
        return model.to_onnx(X=X, name=name, target_opset=target_opset)
    if name is None:
        name = "ONNX(%s)" % model.__class__.__name__
    initial_types = guess_initial_types(X, initial_types)
    if verbose >= 1:
        print("[to_onnx] initial_types=%r" % initial_types)
    return convert_sklearn(model, initial_types=initial_types,
                           target_opset=target_opset,
                           name=name, options=options,
                           white_op=white_op, black_op=black_op,
                           final_types=final_types, dtype=dtype,
                           verbose=verbose, naming=naming,
                           model_optim=model_optim)


def wrap_as_onnx_mixin(model, target_opset=None):
    """
    Combines a *scikit-learn* class with :class:`OnnxOperatorMixin`
    which produces a new object which combines *scikit-learn* API
    and *OnnxOperatorMixin* API.
    """
    from .algebra.sklearn_ops import find_class
    cl = find_class(model.__class__)
    if "automation" in str(cl):
        raise RuntimeError("Wrong class name '{}'.".format(cl))
    state = model.__getstate__()
    obj = object.__new__(cl)
    obj.__setstate__(state)
    obj.op_version = target_opset
    return obj
